"""
`Construct` module
==================

The `construct` module is used to construct all objects that 
form the inputs to calculations in openpile. 


These objects include:

- Pile 
  - PileMaterial
  - PileSection
- SoilProfile
- Layer
- Model

**Usage**

>>> from openpile.construct import Pile, SoilProfile, Layer, Model

"""


import math as m
import pandas as pd
import numpy as np
import warnings
import matplotlib.pyplot as plt

import openpile.utils.graphics as graphics

from openpile.materials import PileMaterial
import openpile.materials as materials
from openpile.core import misc, _model_build
from openpile.soilmodels import LateralModel, AxialModel
from openpile.core.misc import generate_color_string
from openpile.core._model_build import (
    check_springs,
    get_soil_properties,
    apply_bc,
    validate_bc,
    parameter2elements,
)

from abc import ABC, abstractmethod
from typing import List, Dict, Optional, Union
from typing_extensions import Literal, Annotated, Optional
from pydantic import (
    BaseModel,
    ConfigDict,
    InstanceOf,
    Field,
    model_validator,
    computed_field,
)
from functools import cached_property

from pydantic import (
    BaseModel,
    Field,
    model_validator,
)


class AbstractPile(BaseModel, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")


class AbstractLayer(BaseModel, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")


class AbstractSoilProfile(BaseModel, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")


class AbstractModel(BaseModel, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")


class PileSection(BaseModel, ABC):
    """
    An abstract Pile Segment is a section of a pile.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    @property
    @abstractmethod
    def top_elevation(self) -> float:
        pass

    @property
    @abstractmethod
    def bottom_elevation(self) -> float:
        pass

    @property
    @abstractmethod
    def footprint(self) -> float:
        pass

    @property
    @abstractmethod
    def length(self) -> float:
        pass

    @property
    @abstractmethod
    def area(self) -> float:
        pass

    @property
    @abstractmethod
    def entrapped_area(self) -> float:
        pass

    @property
    @abstractmethod
    def outer_perimeter(self) -> float:
        pass

    @property
    @abstractmethod
    def inner_perimeter(self) -> float:
        pass

    @abstractmethod
    def get_volume(self, length) -> float:
        pass

    @abstractmethod
    def get_entrapped_volume(self, length) -> float:
        pass

    @property
    @abstractmethod
    def width(self) -> float:
        pass

    @property
    @abstractmethod
    def second_moment_of_area(self) -> float:
        pass


class CircularPileSection(PileSection):
    """A circular section of a pile.

    Parameters
    ----------
    top : float
        the top elevation of the circular section, in meters
    bottom : float
        the bottom elevation of the circular section, in meters
    diameter : float
        the diameter of the circular section, in meters
    thickness : Optional[float], optional
        the wall thickness of the circular section if the section is hollow, in meters,
        by default None which means the section is solid.

    """

    top: float
    bottom: float
    diameter: Annotated[float, Field(gt=0)]
    thickness: Optional[Annotated[float, Field(gt=0)]] = None

    @model_validator(mode="after")
    def get_proper_thickness(self):
        if self.thickness is None:
            self.thickness = self.diameter / 2
        return self

    @model_validator(mode="after")
    def check_elevations(self):
        if self.bottom >= self.top:
            raise ValueError(
                f"Bottom elevation ({self.bottom}) must be less than top elevation ({self.top})."
            )
        return self

    @property
    def top_elevation(self) -> float:
        return self.top

    @property
    def bottom_elevation(self) -> float:
        return self.bottom

    @property
    def length(self) -> float:
        return self.top - self.bottom

    @property
    def area(self) -> float:
        return (self.diameter**2 - (self.diameter - 2 * self.thickness) ** 2) * m.pi / 4

    @property
    def entrapped_area(self) -> float:
        return (self.diameter - 2 * self.thickness) ** 2 * m.pi / 4

    @property
    def outer_perimeter(self) -> float:
        return self.diameter * m.pi

    @property
    def inner_perimeter(self) -> float:
        return (self.diameter - 2 * self.thickness) * m.pi

    @property
    def footprint(self) -> float:
        return self.diameter**2 * m.pi / 4

    @property
    def width(self) -> float:
        return self.diameter

    @property
    def second_moment_of_area(self) -> float:
        return (self.diameter**4 - (self.diameter - 2 * self.thickness) ** 4) * m.pi / 64

    def get_volume(self, length) -> float:
        return self.area * length

    def get_entrapped_volume(self, length) -> float:
        return length * (self.diameter - 2 * self.thickness) ** 2 * m.pi / 4


class Pile(AbstractPile):
    #: name of the pile
    name: str
    #: There can be as many sections as needed by the user. The length of the listsdictates the number of pile sections.
    sections: List[InstanceOf[PileSection]]
    #: select the type of material the pile is made of, can be of ('Steel', 'Concrete') or a material created from openpile.materials.PileMaterial.custom()
    material: Union[Literal["Steel", "Concrete"], PileMaterial]
    """
    A class to create the pil.e.

    Parameters
    ----------
    name : str
        Pile/Structure's name.
    sections : List[PileSection]
        argument that stores the relevant data of each pile segment. numbering of sections is made from uppermost elevation and 0-indexed.
    material : Literal["Steel",]
        material the pile is made of. by default "Steel"


    Example
    -------

    >>> from openpile.construct import Pile, CircularPileSection
    >>> # Create a pile instance with two sections of respectively 10m and 30m length.
    >>> pile = Pile(name = "",
    ...         material='Steel',
    ...         sections=[
    ...             CircularPileSection(
    ...                 top=0, 
    ...                 bottom=-10, 
    ...                 diameter=7.5, 
    ...                 thickness=0.07
    ...             ),
    ...             CircularPileSection(
    ...                 top=-10, 
    ...                 bottom=-40, 
    ...                 diameter=7.5, 
    ...                 thickness=0.08
    ...             ),
    ...         ]
    ...     )

    One can also create a pile from other constructors such as: create_tubular(), that creates a ciruclar hollow pile of one unique section.

    >>> from openpile.construct import Pile
    >>> pile = Pile.create_tubular(name = "",
    ...         top_elevation = 0,
    ...         bottom_elevation = -40,
    ...         diameter=7.5,
    ...         wt=0.07,
    ...         )
    """

    # check that we have at least one pile section
    @model_validator(mode="after")
    def sections_must_be_provided(self):
        if len(self.sections) == 0:
            raise ValueError("No pile sections provided.")
        return self

    # check that dict is correctly entered
    @model_validator(mode="after")
    def sections_must_not_overlap(self):
        self.sections = sorted(self.sections, key=lambda x: -x.top_elevation)
        for i, segment in enumerate(self.sections):
            if i == 0:
                pass
            else:
                previous_segment = self.sections[i - 1]
                if segment.top_elevation != previous_segment.bottom_elevation:
                    raise ValueError(
                        f"Pile sections are not consistent. Pile section No. {i} and No. {i-1} do not connect."
                    )
        return self

    # check that dict is correctly entered
    @model_validator(mode="after")
    def check_materials(self):
        if self.material == "Steel":
            self.material = materials.steel
        elif self.material == "Concrete":
            self.material = materials.concrete
        return self

    @property
    def top_elevation(self) -> float:
        return self.sections[0].top_elevation

    @property
    def data(self) -> pd.DataFrame:
        # create pile data used by openpile for mesh and calculations.
        # Create top and bottom elevations
        return pd.DataFrame(
            data={
                "Elevation [m]": [
                    x for x in self.sections for x in [x.top_elevation, x.bottom_elevation]
                ],
                "Width [m]": [x.width for x in self.sections for x in [x, x]],
                "Area [m2]": [x.area for x in self.sections for x in [x, x]],
                "I [m4]": [x.second_moment_of_area for x in self.sections for x in [x, x]],
                "Entrapped Area [m2]": [x.entrapped_area for x in self.sections for x in [x, x]],
                "Outer Perimeter [m]": [x.outer_perimeter for x in self.sections for x in [x, x]],
                "Inner Perimeter [m]": [x.inner_perimeter for x in self.sections for x in [x, x]],
            }
        )

    def __str__(self):
        if self.shape == "Circular":
            return pd.DataFrame(
                data={
                    "Elevation [m]": [
                        x for x in self.sections for x in [x.top_elevation, x.bottom_elevation]
                    ],
                    "Diameter [m]": [x.width for x in self.sections for x in [x, x]],
                    "Wall thickness [m]": [x.thickness for x in self.sections for x in [x, x]],
                    "Area [m2]": [x.area for x in self.sections for x in [x, x]],
                    "I [m4]": [x.second_moment_of_area for x in self.sections for x in [x, x]],
                }
            ).to_string()
        else:
            return pd.DataFrame(
                data={
                    "Elevation [m]": [
                        x for x in self.sections for x in [x.top_elevation, x.bottom_elevation]
                    ],
                    "Width [m]": [x.width for x in self.sections for x in [x, x]],
                    "Area [m2]": [x.area for x in self.sections for x in [x, x]],
                    "I [m4]": [x.second_moment_of_area for x in self.sections for x in [x, x]],
                }
            ).to_string()

    @property
    def bottom_elevation(self) -> float:
        """
        Bottom elevation of the pile [m VREF].
        """
        return self.sections[-1].bottom_elevation

    @property
    def length(self) -> float:
        """
        Pile length [m].
        """
        return self.top_elevation - self.bottom_elevation

    @property
    def volume(self) -> float:
        """
        Pile volume [m3].
        """
        return sum([x.area * x.length for x in self.sections])

    @property
    def weight(self) -> float:
        """
        Pile weight [kN].
        """
        return self.volume * self.material.unitweight

    @property
    def G(self) -> float:
        """
        Shear modulus of the pile material [kPa]. Thie value does not vary across and along the pile.
        """
        return self.material.shear_modulus

    @property
    def E(self) -> float:
        """
        Young modulus of the pile material [kPa]. Thie value does not vary across and along the pile.
        """
        return self.material.young_modulus

    @property
    def tip_area(self) -> float:
        "Sectional area at the bottom of the pile [m2]"
        return self.sections[-1].area

    @property
    def tip_footprint(self) -> float:
        "footprint area at the bottom of the pile [m2]"
        return self.sections[-1].footprint

    @property
    def inner_volume(self) -> float:
        """the inner volume in [m3] of the pile from the model object."""

        z_top = np.array([x.top_elevation for x in self.sections])
        z_bottom = np.array([x.bottom_elevation for x in self.sections])
        L = z_top - z_bottom
        area_inside = np.array([x.entrapped_area for x in self.sections])

        return np.sum(area_inside * L)

    @classmethod
    def create_tubular(
        cls,
        name: str,
        top_elevation: float,
        bottom_elevation: float,
        diameter: float,
        wt: float,
        material: Union[Literal["Steel", "Concrete"], PileMaterial] = "Steel",
    ):
        """A method to simplify the creation of a Pile instance.
        This method creates a circular and hollow pile of constant diameter and wall thickness.

        Parameters
        ----------
        name : str
            Pile/Structure's name.
        top_elevation : float
            top elevation of the pile [m VREF]
        bottom_elevation : float
            bottom elevation of the pile [m VREF]
        diameter : float
            pile diameter [m]
        wt : float
            pile's wall thickness [m]
        material : Literal["Steel", "Concrete"] or an instance of openpile.materials.PileMaterial
            material the pile is made of. by default "Steel"

        Returns
        -------
        openpile.construct.Pile
            a Pile instance.
        """

        obj = cls(
            name=name,
            material=material,
            sections=[
                CircularPileSection(
                    top=top_elevation,
                    bottom=bottom_elevation,
                    diameter=diameter,
                    thickness=wt,
                )
            ],
        )
        return obj

    @property
    def shape(self):
        if all([isinstance(x, CircularPileSection) for x in self.sections]):
            return "Circular"
        else:
            "Undefined"


class Layer(AbstractLayer):
    """A class to create a layer.

    The Layer stores information on the soil parameters of the layer as well
    as the relevant/representative constitutive model (aka. the soil spring).

    Parameters
    ----------
    name : str
        Name of the layer, use for printout.
    top : float
        top elevation of the layer in [m].
    bottom : float
        bottom elevation of the layer in [m].
    weight : float
        total unit weight in [kN/m3], cannot be lower than 10.
    lateral_model : LateralModel
        Lateral soil model of the layer, by default None.
    axial_model : AxialModel
        Axial soil model of the layer, by default None.
    color : str
        soil layer color in HEX format (e.g. '#000000'), by default None.
        If None, the color is generated randomly.


    Example
    -------

    >>> from openpile.construct import Layer
    >>> from openpile.soilmodels import API_clay
    >>> # Create a layer with increasing values of Su and eps50
    >>> layer1 = Layer(name='Soft Clay',
    ...            top=0,
    ...            bottom=-10,
    ...            weight=19,
    ...            lateral_model=API_clay(Su=[30,35], eps50=[0.01, 0.02], kind='static'),
    ...            )
    >>> # show layer
    >>> print(layer1) # doctest: +NORMALIZE_WHITESPACE
    Name: Soft Clay
    Elevation: (0.0) - (-10.0) m
    Weight: 19.0 kN/m3
    Lateral model: 	API clay
        Su = 30.0-35.0 kPa
        eps50 = 0.01-0.02
        static curves
        ext: None
    Axial model: None
    """

    #: name of the layer, use for printout
    name: str
    #: top elevaiton of the layer
    top: float
    #: bottom elevaiton of the layer
    bottom: float
    #: unit weight in kN of the layer
    weight: Annotated[float, Field(gt=10.0)]
    #: Lateral constitutive model of the layer
    lateral_model: Optional[LateralModel] = None
    #: Axial constitutive model of the layer
    axial_model: Optional[AxialModel] = None
    #: Layer's color when plotted
    color: Optional[Annotated[str, Field(min_length=7, max_length=7)]] = None

    def model_post_init(self, *args, **kwargs):
        if self.color is None:
            self.color = generate_color_string("earth")
        return self

    def __str__(self):
        return f"Name: {self.name}\nElevation: ({self.top}) - ({self.bottom}) m\nWeight: {self.weight} kN/m3\nLateral model: {self.lateral_model}\nAxial model: {self.axial_model}"

    @model_validator(mode="after")
    def check_elevations(self):
        if not self.top > self.bottom:
            raise ValueError("Bottom elevation is higher than top elevation")
        else:
            return self


class SoilProfile(AbstractSoilProfile):
    """
    A class to create the soil profile. A soil profile consist of a ground elevation (or top elevation)
    with one or more layers of soil.

    Additionally, a soil profile can include discrete information at given elevation such as CPT
    (Cone Penetration Test) data. Not Implemented yet!

    Parameters
    ----------
    name : str
        Name of the soil profile, used for printout and plots.
    top_elevation : float
        top elevation of the soil profile in [m VREF].
    water_line : float
        elevation of the water table in [m VREF].
    layers : list[Layer]
        list of layers for the soil profile.
    cpt_data : np.ndarray
        cpt data table with
        1st col: elevation [m],
        2nd col: cone resistance [kPa],
        3rd col: sleeve friction [kPa],
        4th col: pore pressure u2 [kPa].

    Example
    -------
    >>> # import objects
    >>> from openpile.construct import SoilProfile, Layer
    >>> from openpile.soilmodels import API_sand, API_clay
    >>> # Create a two-layer soil profile
    >>> sp = SoilProfile(
    ...     name="BH01",
    ...     top_elevation=0,
    ...     water_line=0,
    ...     layers=[
    ...         Layer(
    ...             name='Layer0',
    ...             top=0,
    ...             bottom=-20,
    ...             weight=18,
    ...             lateral_model= API_sand(phi=30, kind='cyclic')
    ...         ),
    ...         Layer( name='Layer1',
    ...                 top=-20,
    ...                 bottom=-40,
    ...                 weight=19,
    ...                 lateral_model= API_clay(Su=50, eps50=0.01, kind='static'),)
    ...     ]
    ... )
    >>> # Check soil profile content
    >>> print(sp) # doctest: +NORMALIZE_WHITESPACE
    Layer 1
    ------------------------------
    Name: Layer0
    Elevation: (0.0) - (-20.0) m
    Weight: 18.0 kN/m3
    Lateral model: 	API sand
        phi = 30.0°
        cyclic curves
        ext: None
    Axial model: None
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Layer 2
    ------------------------------
    Name: Layer1
    Elevation: (-20.0) - (-40.0) m
    Weight: 19.0 kN/m3
    Lateral model: 	API clay
        Su = 50.0 kPa
        eps50 = 0.01
        static curves
        ext: None
    Axial model: None
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    """

    #: name of soil profile / borehole / location
    name: str
    #: top of ground elevation with respect to the model reference elevation datum
    top_elevation: float
    #: water elevation (this can refer to sea elevation of water table)
    water_line: float
    #: soil layers to consider in the soil propfile
    layers: List[Layer]
    #: Cone Penetration Test data with folloeing structure:
    #: 1st col: elevation[m],
    #: 2nd col: cone resistance[kPa],
    #: 3rd col: sleeve friction [kPa]
    #: 4th col: pore pressure u2 [kPa]
    #: (the cpt data outside the soil profile boundaries will be ignored)
    cpt_data: Optional[np.ndarray] = None

    @model_validator(mode="after")
    def check_layers_elevations(self):

        top_elevations = np.array([x.top for x in self.layers], dtype=float)
        bottom_elevations = np.array([x.bottom for x in self.layers], dtype=float)
        idx_sort = np.argsort(top_elevations)

        top_sorted = top_elevations[idx_sort][::-1]
        bottom_sorted = bottom_elevations[idx_sort][::-1]

        # check no overlap
        if top_sorted[0] != self.top_elevation:
            raise ValueError("top_elevation not matching uppermost layer's elevations.")

        for i in range(len(top_sorted) - 1):
            if not m.isclose(top_sorted[i + 1], bottom_sorted[i], abs_tol=0.001):
                raise ValueError("Layers' elevations overlap.")

        return self

    @model_validator(mode="after")
    def check_multipliers_in_lateral_model(self):
        def check_multipliers_callable(multiplier, ground_level, top, bottom, type):
            # if not a float, it must be a callable, then we check for Real Positive float
            if not isinstance(multiplier, float):
                # defines depth below ground to check
                depths = ground_level - np.linspace(start=top, stop=bottom, num=100)
                # check if positive real float is returned
                for depth in depths:
                    result = multiplier(depth)
                    if not isinstance(result, float):
                        TypeError(
                            f"One or more results of the {type}-multiplier callable is not a float"
                        )
                        return None
                    else:
                        if type in ["p", "m"]:
                            if result < 0.0:
                                print(
                                    f"One or more results of the {type}-multiplier callable is negative"
                                )
                                return None
                        elif type in ["y", "t"]:
                            if not result > 0.0:
                                ValueError(
                                    f"One or more results of the {type}-multiplier callable is not strictly positive"
                                )
                                return None

        for layer in self.layers:
            if layer.lateral_model is not None:
                # check p-multipliers
                check_multipliers_callable(
                    layer.lateral_model.p_multiplier,
                    self.top_elevation,
                    layer.top,
                    layer.bottom,
                    "p",
                )
                # check y-multipliers
                check_multipliers_callable(
                    layer.lateral_model.y_multiplier,
                    self.top_elevation,
                    layer.top,
                    layer.bottom,
                    "y",
                )
                # check m-multipliers
                check_multipliers_callable(
                    layer.lateral_model.m_multiplier,
                    self.top_elevation,
                    layer.top,
                    layer.bottom,
                    "m",
                )
                # check t-multipliers
                check_multipliers_callable(
                    layer.lateral_model.t_multiplier,
                    self.top_elevation,
                    layer.top,
                    layer.bottom,
                    "t",
                )

        return self

    def __str__(self):
        """List all layers in table-like format"""
        out = ""
        i = 0
        for layer in self.layers:
            i += 1
            out += f"Layer {i}\n" + "-" * 30 + "\n"
            out += f"{layer}\n" + "~" * 30 + "\n"
        return out

    @property
    def bottom_elevation(self) -> float:
        """
        Bottom elevation of the soil profile [m VREF].
        """
        return self.top_elevation - sum([abs(x.top - x.bottom) for x in self.layers])

    def plot(self, ax=None):
        """Creates a plot illustrating the stratigraphy.

        Parameters
        ----------
        ax : axis handle from matplotlib figure, optional
            if None, a new axis handle is created

        Returns
        -------
        matplotlib.pyplot.figure
            only return the object if assign=True

        Example
        -------

        .. plot::
            :context: reset
            :include-source: False

            from openpile.construct import Pile, CircularPileSection
            # Create a pile instance with two sections of respectively 10m and 30m length.
            p = Pile(name = "",
                    material='Steel',
                    sections=[
                        CircularPileSection(
                            top=0,
                            bottom=-10,
                            diameter=7.5,
                            thickness=0.07
                        ),
                        CircularPileSection(
                            top=-10,
                            bottom=-40,
                            diameter=7.5,
                            thickness=0.08
                        ),
                    ]
            )

            from openpile.construct import SoilProfile, Layer

            # save the figure generated by SoilProfile.plot()
            sp = SoilProfile(
                name="<soil profile name>",
                top_elevation=0,
                water_line=8.0,
                layers=[
                    Layer(name="<layer name>", top=0, bottom=-4, weight=18),
                    Layer(name="<layer name>", top=-4, bottom=-8, weight=18),
                    Layer(name="<layer name>", top=-8, bottom=-20, weight=18),
                    Layer(name="<layer name>", top=-20, bottom=-40, weight=18),
                ],
            )
            sp.plot()
        """
        ax = graphics.soil_plot(self, ax=ax)


class BoundaryFixation(BaseModel):
    """
    A class to create a boundary condition where support is fixed.

    Parameters
    ----------
    elevation : str
        Elevation of the boundary condition [m VREF]
    x : bool
        Fix the boundary condition in the x-direction
    y : bool
        Fix the boundary condition in the y-direction
    z : bool
        Fix the boundary condition in the z-direction
    """

    elevation: float
    x: Optional[bool] = None
    y: Optional[bool] = None
    z: Optional[bool] = None


class BoundaryDisplacement(BaseModel):
    """
    A class to create a boundary condition where displacement is given.

    Parameters
    ----------
    elevation : str
        Elevation of the boundary condition [m VREF]
    x : float
        Apply displacement in the x-direction [m]
    y : float
        Apply displacement in the y-direction [m]
    z : float
        Apply displacement in the z-direction [m]
    """

    elevation: float
    x: Optional[float] = None
    y: Optional[float] = None
    z: Optional[float] = None


class BoundaryForce(BaseModel):
    """
    A class to create a boundary condition where force is given.

    Parameters
    ----------
    elevation : str
        Elevation of the boundary condition [m VREF]
    x : float
        Apply force in the x-direction [kN]
    y : float
        Apply force in the y-direction [kN]
    z : float
        Apply force in the z-direction [kN]
    """

    elevation: float
    x: Optional[float] = None
    y: Optional[float] = None
    z: Optional[float] = None


class Model(AbstractModel):
    """
    A class to create a Model.

    A Model is constructed based on the pile geometry/data primarily.
    Additionally, a soil profile can be fed to the Model, and soil springs can be created.

    Parameters
    ----------
    name : str
        Name of the model
    pile : Pile
        Pile instance to be included in the model.
    boundary_conditions : list[BoundaryFix, BoundaryDisp, BoundaryForce], optional
        list of boundary conditions to be included in the model, by default None.
        Boundary conditions can be added when instantiating the model with Boundary... objects or via the methods:
        `.set_pointload()`, `.set_pointdisplacement()`, `.set_support()`
    soil : Optional[SoilProfile], optional
        SoilProfile instance, by default None.
    element_type : str, optional
        can be of ['EulerBernoulli','Timoshenko'], by default 'Timoshenko'.
    x2mesh : List[float], optional
        additional elevations to be included in the mesh, by default none.
    coarseness : float, optional
        maximum distance in meters between two nodes of the mesh, by default 0.5. A value lower than 0.01 is not accepted due to computational purposes.
    distributed_lateral : bool, optional
        include distributed lateral springs, by default True.
    distributed_moment : bool, optional
        include distributed moment springs, by default False.
    base_shear : bool, optional
        include lateral spring at pile toe, by default False.
    base_moment : bool, optional
        include moment spring at pile toe, by default False.
    distributed_axial : bool, optional
        include distributed axial springs, by default True.
    base_axial : bool, optional
        include base axial springs, by default True.
    plugging : bool, optional
        whether the pile is plugged or unplugged, by default False.


    Example
    -------

    >>> from openpile.construct import Pile, Model, Layer
    >>> from openpile.soilmodels import API_sand
    >>> # create pile
    ... p = Pile(name = "",
    ...          material='Steel',
    ...          sections=[
    ...             CircularPileSection(
    ...                 top=0,
    ...                 bottom=-10,
    ...                 diameter=7.5,
    ...                 thickness=0.07
    ...             ),
    ...             CircularPileSection(
    ...                 top=-10,
    ...                 bottom=-40,
    ...                 diameter=7.5,
    ...                 thickness=0.08
    ...             ),
    ...          ]
    ...     )

    >>> # Create Soil Profile
    >>> sp = SoilProfile(
    ... 	name="BH01",
    ... 	top_elevation=0,
    ... 	water_line=0,
    ... 	layers=[
    ... 		Layer(
    ... 			name='Layer0',
    ... 			top=0,
    ... 			bottom=-40,
    ... 			weight=18,
    ... 			lateral_model= API_sand(phi=30, kind='cyclic')
    ... 		),
    ... 	]
    ... )
    >>> # Create Model
    >>> M = Model(name="Example", pile=p, soil=sp)
    >>> # create Model without soil maximum 5 metres apart.
    >>> Model_without_soil = Model(name = "Example without soil", pile=p, coarseness=5)
    >>> # create Model with nodes maximum 1 metre apart with soil profile
    >>> Model_with_soil = Model(name = "Example with soil", pile=p, soil=sp, coarseness=1)
    """

    #: model name
    name: str
    #: pile instance that the Model should consider
    pile: Pile
    #: boundary conditions of the model
    boundary_conditions: List[Union[BoundaryFixation, BoundaryForce, BoundaryDisplacement]] = Field(
        default_factory=list
    )
    #: soil profile instance that the Model should consider
    soil: Optional[SoilProfile] = None
    #: type of beam elements
    element_type: Literal["Timoshenko", "EulerBernoulli"] = "Timoshenko"
    #: x coordinates values to mesh as nodes
    x2mesh: List[float] = Field(default_factory=list)
    #: mesh coarseness, represent the maximum accepted length of elements
    coarseness: Annotated[float, Field(ge=0.01)] = 0.5
    #: whether to include p-y springs in the calculations
    distributed_lateral: bool = True
    #: whether to include m-t springs in the calculations
    distributed_moment: bool = True
    #: whether to include Hb-y spring in the calculations
    base_shear: bool = True
    #: whether to include Mb-t spring in the calculations
    base_moment: bool = True
    #: whether to include t-z springs in the calculations
    distributed_axial: bool = True
    #: whether to include Q-z spring in the calculations
    base_axial: bool = True
    #: plugging
    plugging: bool = None

    @model_validator(mode="after")
    def max_coarseness(self):
        multiplier = 5
        if self.pile.length >= multiplier * self.coarseness:
            return self
        else:
            raise ValueError(
                f"the coarseness factor is too high, please decrease it to at least a value of {np.floor(self.pile.length*100/multiplier)/100}"
            )

    @model_validator(mode="after")
    def soil_and_pile_bottom_elevation_match(self):
        if self.soil is None:
            pass
        else:
            if self.pile.bottom_elevation < self.soil.bottom_elevation:
                raise ValueError("The pile ends deeper than the soil profile.")
        return self

    @model_validator(mode="after")
    def bc_validation(self):
        # validate_bc(self.boundary_conditions, BoundaryDisplacement)
        # validate_bc(self.boundary_conditions, BoundaryForce)
        # validate_bc(self.boundary_conditions, BoundaryFixation)
        return self

    @computed_field
    # @cached_property
    def soil_properties(self) -> Union[Dict[str, np.ndarray], None]:
        return get_soil_properties(self.pile, self.soil, self.x2mesh, self.coarseness)

    @computed_field
    # @cached_property
    def element_properties(self) -> Dict[str, np.ndarray]:
        # creates element structural properties
        # merge Pile.data and coordinates
        element_properties = pd.merge_asof(
            left=self.element_coordinates.sort_values(by=["z_top [m]"]),
            right=self.pile.data.sort_values(by=["Elevation [m]"]),
            left_on="z_top [m]",
            right_on="Elevation [m]",
            direction="forward",
        ).sort_values(by=["z_top [m]"], ascending=False)
        # add young modulus to data
        element_properties["E [kPa]"] = self.pile.material.young_modulus
        # delete Elevation [m] column
        element_properties.drop("Elevation [m]", inplace=True, axis=1)
        # reset index
        element_properties.reset_index(inplace=True, drop=True)

        return element_properties

    @computed_field
    @cached_property
    def nodes_coordinates(self) -> Dict[str, np.ndarray]:
        return _model_build.get_coordinates(self.pile, self.soil, self.x2mesh, self.coarseness)[0]

    @computed_field
    @cached_property
    def element_coordinates(self) -> Dict[str, np.ndarray]:
        return _model_build.get_coordinates(self.pile, self.soil, self.x2mesh, self.coarseness)[1]

    @computed_field
    @property
    def global_forces(self) -> Dict[str, np.ndarray]:

        validate_bc(self.boundary_conditions, BoundaryForce)

        # Initialise nodal global forces with link to nodes_coordinates (used for force-driven calcs)
        df = self.nodes_coordinates.copy()
        df["Pz [kN]"] = 0
        df["Py [kN]"] = 0
        df["Mx [kNm]"] = 0

        nodes_elevations = df["z [m]"].values

        df["Pz [kN]"], df["Py [kN]"], df["Mx [kNm]"] = apply_bc(
            nodes_elevations,
            df["Pz [kN]"].values,
            df["Py [kN]"].values,
            df["Mx [kNm]"].values,
            self.boundary_conditions,
            BoundaryForce,
            "Load",
        )

        return df

    @computed_field
    @property
    def global_disp(self) -> Dict[str, np.ndarray]:

        validate_bc(self.boundary_conditions, BoundaryDisplacement)

        # Initialise nodal global displacement with link to nodes_coordinates (used for displacement-driven calcs)
        df = self.nodes_coordinates.copy()
        df["Tz [m]"] = 0.0
        df["Ty [m]"] = 0.0
        df["Rx [rad]"] = 0.0

        nodes_elevations = df["z [m]"].values

        df["Tz [m]"], df["Ty [m]"], df["Rx [rad]"] = apply_bc(
            nodes_elevations,
            df["Tz [m]"].values,
            df["Ty [m]"].values,
            df["Rx [rad]"].values,
            self.boundary_conditions,
            BoundaryDisplacement,
            "Displacement",
        )

        return df

    @computed_field
    @property
    def global_restrained(self) -> Dict[str, np.ndarray]:

        validate_bc(self.boundary_conditions, BoundaryFixation)

        # Initialise nodal global support with link to nodes_coordinates (used for defining boundary conditions)
        df = self.nodes_coordinates.copy()
        df["Tz"] = False
        df["Ty"] = False
        df["Rx"] = False

        nodes_elevations = df["z [m]"].values

        df["Tz"], df["Ty"], df["Rx"] = apply_bc(
            nodes_elevations,
            df["Tz"].values,
            df["Ty"].values,
            df["Rx"].values,
            self.boundary_conditions,
            BoundaryFixation,
            "Fixity",
        )
        return df

    @property
    def element_number(self) -> int:
        return self.element_properties.shape[0]

    def model_post_init(self, *args, **kwargs):
        def create_springs() -> np.ndarray:
            # dim of springs
            spring_dim = 15
            # springs dim for axial
            tz_springs_dim = 15
            qz_spring_dim = 15

            # Allocate array
            py = np.zeros(shape=(self.element_number, 2, 2, spring_dim), dtype=np.float32)
            mt = np.zeros(
                shape=(self.element_number, 2, 2, spring_dim, spring_dim), dtype=np.float32
            )
            Hb = np.zeros(shape=(1, 1, 2, spring_dim), dtype=np.float32)
            Mb = np.zeros(shape=(1, 1, 2, spring_dim), dtype=np.float32)

            # allocate array for axial springs
            tz = np.zeros(shape=(self.element_number, 2, 2, tz_springs_dim), dtype=np.float32)
            qz = np.zeros(shape=(1, 1, 2, qz_spring_dim), dtype=np.float32)

            soil_prop = self.soil_properties

            # fill in spring for each element
            for layer in self.soil.layers:
                elements_for_layer = soil_prop.loc[
                    (soil_prop["z_top [m]"] <= layer.top)
                    & (soil_prop["z_bottom [m]"] >= layer.bottom)
                ].index

                for i in elements_for_layer:
                    # Set local layer parameters for each element of the layer
                    # vertical effective stress
                    sig_v = soil_prop[["sigma_v top [kPa]", "sigma_v bottom [kPa]"]].iloc[i].values
                    # elevation
                    elevation = soil_prop[["z_top [m]", "z_bottom [m]"]].iloc[i].values
                    # depth from ground
                    depth_from_ground = (
                        (soil_prop[["zg_top [m]", "zg_bottom [m]"]].iloc[i]).abs().values
                    )
                    # pile width
                    pile_width = self.element_properties["Width [m]"].iloc[i]
                    perimeter_out = self.element_properties["Outer Perimeter [m]"].iloc[i]
                    perimeter_in = self.element_properties["Outer Perimeter [m]"].iloc[i]

                    sig_v_tip = soil_prop["sigma_v bottom [kPa]"].iloc[-1]

                    # t-z curves
                    if layer.axial_model is not None:

                        if self.distributed_axial:  # True if tz spring function exist
                            # calculate springs (top and bottom) for each element
                            for j in [0, 1]:
                                (tz[i, j, 1], tz[i, j, 0]) = layer.axial_model.tz_spring_fct(
                                    sig=sig_v[j],
                                    X=depth_from_ground[j],
                                    circumference_out=perimeter_out,
                                    circumference_in=perimeter_in,
                                    layer_height=(layer.top - layer.bottom),
                                    depth_from_top_of_layer=(layer.top - elevation[j]),
                                    D=pile_width,
                                    # TODO add wall thickness for CPT methods?
                                    L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                    below_water_table=elevation[j] <= self.soil.water_line,
                                    output_length=tz_springs_dim,
                                )

                        if (
                            layer.top >= self.pile.bottom_elevation
                            and layer.bottom <= self.pile.bottom_elevation
                            and self.base_axial
                        ):
                            # calculate Qz spring
                            (qz[0, 0, 1], qz[0, 0, 0]) = layer.axial_model.Qz_spring_fct(
                                sig=sig_v_tip,
                                tip_area=self.pile.tip_area,
                                footprint=self.pile.tip_footprint,
                                layer_height=(layer.top - layer.bottom),
                                depth_from_top_of_layer=(layer.top - self.pile.bottom_elevation),
                                D=pile_width,
                                L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                below_water_table=self.pile.bottom_elevation
                                <= self.soil.water_line,
                                output_length=qz_spring_dim,
                            )

                    # py curve
                    if layer.lateral_model is not None:

                        # p-y curves
                        if (
                            layer.lateral_model.spring_signature[0] and self.distributed_lateral
                        ):  # True if py spring function exist

                            # calculate springs (top and) for each element
                            for j in [0, 1]:
                                (py[i, j, 1], py[i, j, 0]) = layer.lateral_model.py_spring_fct(
                                    sig=sig_v[j],
                                    X=depth_from_ground[j],
                                    layer_height=(layer.top - layer.bottom),
                                    depth_from_top_of_layer=(layer.top - elevation[j]),
                                    D=pile_width,
                                    L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                    below_water_table=elevation[j] <= self.soil.water_line,
                                    output_length=spring_dim,
                                )

                        if (
                            layer.lateral_model.spring_signature[2] and self.distributed_moment
                        ):  # True if mt spring function exist

                            # calculate springs (top and) for each element
                            for j in [0, 1]:
                                (mt[i, j, 1], mt[i, j, 0]) = layer.lateral_model.mt_spring_fct(
                                    sig=sig_v[j],
                                    X=depth_from_ground[j],
                                    layer_height=(layer.top - layer.bottom),
                                    depth_from_top_of_layer=(layer.top - elevation[j]),
                                    D=pile_width,
                                    L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                    below_water_table=elevation[j] <= self.soil.water_line,
                                    output_length=spring_dim,
                                )

                        # check if pile tip is within layer
                        if (
                            layer.top >= self.pile.bottom_elevation
                            and layer.bottom <= self.pile.bottom_elevation
                        ):

                            # Hb curve
                            if layer.lateral_model.spring_signature[1] and self.base_shear:

                                # calculate Hb spring
                                (Hb[0, 0, 1], Hb[0, 0, 0]) = layer.lateral_model.Hb_spring_fct(
                                    sig=sig_v_tip,
                                    X=(self.soil.top_elevation - self.soil.bottom_elevation),
                                    layer_height=(layer.top - layer.bottom),
                                    depth_from_top_of_layer=(
                                        layer.top - self.pile.bottom_elevation
                                    ),
                                    D=pile_width,
                                    L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                    below_water_table=self.pile.bottom_elevation
                                    <= self.soil.water_line,
                                    output_length=spring_dim,
                                )

                            # Mb curve
                            if layer.lateral_model.spring_signature[3] and self.base_moment:

                                (Mb[0, 0, 1], Mb[0, 0, 0]) = layer.lateral_model.Mb_spring_fct(
                                    sig=sig_v_tip,
                                    X=(self.soil.top_elevation - self.soil.bottom_elevation),
                                    layer_height=(layer.top - layer.bottom),
                                    depth_from_top_of_layer=(
                                        layer.top - self.pile.bottom_elevation
                                    ),
                                    D=pile_width,
                                    L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                    below_water_table=self.pile.bottom_elevation
                                    <= self.soil.water_line,
                                    output_length=spring_dim,
                                )

            # ensure springs are oriented correctly with respect to z-axis
            # going down is compression and should be negative in "z" values
            tz[:, :, :] = tz[:, :, :] * (-1)
            qz[:, :, :] = qz[:, :, :] * (-1)

            if check_springs(py):
                print("py springs have negative or NaN values.")
                print(
                    """if using PISA type springs, this can be due to parameters behind out of the parameter space.
                Please check that: 2 < L/D < 6.
                """
                )
            if check_springs(mt):
                print("mt springs have negative or NaN values.")
                print(
                    """if using PISA type springs, this can be due to parameters behind out of the parameter space.
                Please check that: 2 < L/D < 6.
                """
                )
            if check_springs(Hb):
                print("Hb spring has negative or NaN values.")
                print(
                    """if using PISA type springs, this can be due to parameters behind out of the parameter space.
                Please check that: 2 < L/D < 6.
                """
                )
            if check_springs(Mb):
                print("Mb spring has negative or NaN values.")
                print(
                    """if using PISA type springs, this can be due to parameters behind out of the parameter space.
                Please check that: 2 < L/D < 6.
                """
                )

            return py, mt, Hb, Mb, tz, qz

        if self.soil is not None:
            # Create arrays of springs
            (
                self._py_springs,
                self._mt_springs,
                self._Hb_spring,
                self._Mb_spring,
                self._tz_springs,
                self._qz_spring,
            ) = create_springs()

    @property
    def embedment(self) -> float:
        """Pile embedment length [m].

        Returns
        -------
        float (or None if no SoilProfile is present)
            Pile embedment
        """
        if self.soil is None:
            return None
        else:
            return self.soil.top_elevation - self.pile.bottom_elevation

    @property
    def top(self) -> float:
        """top elevation of the model [m].

        Returns
        -------
        float
        """
        if self.soil is None:
            return self.pile.top_elevation
        else:
            return max(self.pile.top_elevation, self.soil.top_elevation, self.soil.water_line)

    @property
    def bottom(self) -> float:
        """bottom elevation of the model [m].

        Returns
        -------
        float
        """

        if self.soil is None:
            return self.pile.bottom_elevation
        else:
            return min(self.pile.bottom_elevation, self.soil.bottom_elevation)

    @property
    def effective_pile_weight(self) -> float:
        if self.soil is not None:
            submerged_element = self.element_properties["z_top [m]"].values <= self.soil.water_line

            elem_z_top = self.element_properties["z_top [m]"].values
            elem_z_bottom = self.element_properties["z_bottom [m]"].values
            V = (elem_z_top - elem_z_bottom) * parameter2elements(
                self.pile.sections, lambda x: x.area, elem_z_top, elem_z_bottom
            )
            W = np.zeros(shape=V.shape)
            W[submerged_element] = V[submerged_element] * (self.pile.material.unitweight - 10)
            W[~submerged_element] = V[~submerged_element] * (self.pile.material.unitweight)
            return W.sum()

        else:
            raise Exception(
                "Model must be linked to a soil profile, use `openpile.construct.Pile.weight instead.`"
            )

    @property
    def shaft_resistance(self) -> float:
        "the shaft resistances [kN] in compression and tension respectively calculated from the provided axial models along the pile."

        if self.soil is None:
            raise Exception("Model must be linked to a soil profile with provided axial models.")
        else:
            # influence zones
            influence = np.abs(np.gradient(self.nodes_coordinates["z [m]"].values))
            influence = influence / 2
            influence

            # Shaft resistance calc compression
            comp = np.sum(
                np.abs(
                    np.min(self._tz_springs[:, 0, 0, :], axis=1) * influence[:-1]
                    + np.min(self._tz_springs[:, 1, 0, :], axis=1) * influence[1:]
                )
            )
            # Shaft resistance calc tension
            tens = np.sum(
                np.abs(
                    np.max(self._tz_springs[:, 0, 0, :], axis=1) * influence[:-1]
                    + np.max(self._tz_springs[:, 1, 0, :], axis=1) * influence[1:]
                )
            )

            return comp, tens

    @property
    def tip_resistance(self) -> float:
        "the end bearing resistance [kN] calculated from the provided axial model at tip elevation"
        if self.soil is not None:
            return float(np.abs(np.min(self._qz_spring[0, 0, 0, :])))
        else:
            raise Exception("Model must be linked to a soil profile with provided axial models.")

    @property
    def entrapped_soil_weight(self) -> float:
        """calculates total weight of soil inside the pile. (Unit: kN)"""

        if self.soil is not None:

            # weight water in kN/m3
            uw_water = 10

            # soil volume
            elem_z_top = self.element_properties["z_top [m]"].values
            elem_z_bottom = self.element_properties["z_bottom [m]"].values
            L = elem_z_top - elem_z_bottom
            area_inside = parameter2elements(
                self.pile.sections, lambda x: x.entrapped_area, elem_z_top, elem_z_bottom
            )
            Vi = area_inside * L
            # element mid-point elevation
            elevation = 0.5 * (
                self.soil_properties["z_top [m]"] + self.soil_properties["z_bottom [m]"]
            )
            # soil weight for each element where we have soil and pile
            elem_number = int(self.element_properties.shape[0])
            element_sw = np.zeros(elem_number)

            for layer in self.soil.layers:
                elements_for_layer = self.soil_properties.loc[
                    (self.soil_properties["z_top [m]"] <= layer.top)
                    & (self.soil_properties["z_bottom [m]"] >= layer.bottom)
                ].index

                # Set local layer parameters for each element of the layer
                for i in elements_for_layer:
                    # Calculate inner soil weight
                    element_sw[i] = (
                        layer.weight * Vi[i]
                        if elem_z_top[i] > self.soil.water_line
                        else (layer.weight - uw_water) * Vi[i]
                    )
            return element_sw.sum()
        else:
            raise Exception(
                "Model must be linked to a soil profile, the argument `soil` must be provided."
            )

    def get_structural_properties(self) -> pd.DataFrame:
        """
        Returns a table with the structural properties of the pile sections.
        """
        try:
            return self.element_properties
        except AttributeError:
            print("Data not found. Please create Model with the Model.create() method.")
        except Exception as e:
            print(e)

    def get_soil_properties(self) -> pd.DataFrame:
        """
        Returns a table with the soil main properties and soil models of each element.
        """
        try:
            return self.soil_properties
        except AttributeError:
            print("Data not found. Please create Model with the Model.create() method.")
        except Exception as e:
            print(e)

    def _update_bc(self, elevation, x, y, z, BC_class):
        # list existing
        existing_bcs = [
            bc
            for bc in self.boundary_conditions
            if bc.elevation == elevation and isinstance(bc, BC_class)
        ]

        # check if elevation already exists for the provided axis
        if len(existing_bcs) == 1:
            # check if elevation already exists for the provided axis and modify boundary condition object
            if x is not None:
                existing_bcs[0].x = x
            if y is not None:
                existing_bcs[0].y = y
            if z is not None:
                existing_bcs[0].z = z
        else:
            self.boundary_conditions.append(BC_class(elevation=elevation, x=x, y=y, z=z))

    def set_pointload(
        self,
        *,
        elevation: float = 0.0,
        Py: float = None,
        Pz: float = None,
        Mx: float = None,
    ):
        """
        Defines the point load(s) at a given elevation.

        .. note:
            If run several times at the same elevation, the loads along relevant axes are overwritten by the last command.

        Parameters
        ----------
        elevation : float, optional
            the elevation must match the elevation of a node, by default 0.0.
        Py : float, optional
            Shear force in kN, by default None.
        Pz : float, optional
            Normal force in kN, by default None.
        Mx : float, optional
            Bending moment in kNm, by default None.
        """
        self._update_bc(elevation, z=Pz, y=Py, x=Mx, BC_class=BoundaryForce)

    def set_pointdisplacement(
        self,
        elevation: float = 0.0,
        Ty: float = None,
        Tz: float = None,
        Rx: float = None,
    ):
        """
        Defines the displacement at a given elevation.

        .. note::
            for defining supports, this function should not be used, rather use `.set_support()`.

        .. note:
            If run several times at the same elevation, the displacements along relevant axes are overwritten by the last command.

        Parameters
        ----------
        elevation : float, optional
            the elevation must match the elevation of a node, by default 0.0.
        Ty : float, optional
            Translation along y-axis, by default None.
        Tz : float, optional
            Translation along z-axis, by default None.
        Rx : float, optional
            Rotation around x-axis, by default None.
        """
        self._update_bc(elevation, z=Tz, y=Ty, x=Rx, BC_class=BoundaryDisplacement)

    def set_support(
        self,
        elevation: float = 0.0,
        Ty: bool = None,
        Tz: bool = None,
        Rx: bool = None,
    ):
        """
        Defines the supports at a given elevation. If True, the relevant degree of freedom is restrained.

        .. note:
            If run several times at the same elevation, the support along relevant axes are overwritten by the last command.


        Parameters
        ----------
        elevation : float, optional
            the elevation must match the elevation of a node, by default 0.0.
        Ty : bool, optional
            Translation along y-axis, by default None.
        Tz : bool, optional
            Translation along z-axis, by default None.
        Rx : bool, optional
            Rotation around x-axis, by default None.
        """
        self._update_bc(elevation, z=Tz, y=Ty, x=Rx, BC_class=BoundaryFixation)

    def get_distributed_axial_springs(self, kind: str = "lumped") -> pd.DataFrame:
        """Table with t-z springs computed for the given Model with t-value [kN/m] and y-value [m].

        Posible to extract the springs as typical structural springs (which are also the raw
        springs used in the model) or element level (i.e. top and bottom springs at each element)

        Parameters
        ----------
        kind : str
            can be of ("lumped", "distributed").

        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with t-z springs
        """
        if self.soil is None:
            return None
        else:
            if kind == "distributed":
                return misc.get_distributed_soil_springs(
                    springs=self._tz_springs,
                    elevations=self.nodes_coordinates["z [m]"].values,
                    kind="t-z",
                )
            elif kind == "lumped":
                return misc.get_lumped_soil_springs(
                    springs=self._tz_springs,
                    elevations=self.nodes_coordinates["z [m]"].values,
                    kind="t-z",
                )
            else:
                return None

    def get_distributed_lateral_springs(self, kind: str = "lumped") -> pd.DataFrame:
        """Table with p-y springs computed for the given Model with p-value [kN/m] and y-value [m].

        Posible to extract the springs as typical structural springs (which are also the raw
        springs used in the model) or element level (i.e. top and bottom springs at each element)

        Parameters
        ----------
        kind : str
            can be of ("lumped", "distributed").

        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with p-y springs
        """
        if self.soil is None:
            return None
        else:
            if kind == "distributed":
                return misc.get_distributed_soil_springs(
                    springs=self._py_springs,
                    elevations=self.nodes_coordinates["z [m]"].values,
                    kind="p-y",
                )
            elif kind == "lumped":
                return misc.get_lumped_soil_springs(
                    springs=self._py_springs,
                    elevations=self.nodes_coordinates["z [m]"].values,
                    kind="p-y",
                )
            else:
                return None

    def get_distributed_rotational_springs(self, kind: str = "lumped") -> pd.DataFrame:
        """Table with m-t (rotational) springs computed for the given Model with m-value [kNm] and t-value [radians]

        Posible to extract the springs as typical structural springs (which are also the raw
        springs used in the model) or element level (i.e. top and bottom springs at each element)

        Parameters
        ----------
        kind : str
            can be of ("lumped", "distributed").

        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with m-t springs.
        """
        if self.soil is None:
            return None
        else:
            if kind == "distributed":
                return misc.get_soil_springs(
                    springs=self._mt_springs,
                    elevations=self.nodes_coordinates["z [m]"].values,
                    kind="m-t",
                )
            elif kind == "lumped":
                return misc.get_lumped_soil_springs(
                    springs=self._mt_springs,
                    elevations=self.nodes_coordinates["z [m]"].values,
                    kind="m-t",
                )
            else:
                return None

    def get_base_shear_spring(self) -> pd.DataFrame:
        """Table with Hb (base shear) spring computed for the given Model with Hb-value [kN] and y-value [m].

        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with Hb spring.
        """
        if self.soil is None:
            return None
        else:
            spring_dim = self._Hb_spring.shape[-1]

            column_values_spring = [f"VAL {i}" for i in range(spring_dim)]

            df = pd.DataFrame(
                data={
                    "Node no.": [self.element_number + 1] * 2,
                    "Elevation [m]": [self.pile.bottom_elevation] * 2,
                }
            )
            df["type"] = ["Hb", "y"]
            df[column_values_spring] = self._Hb_spring.reshape(2, spring_dim)

            return df

    def get_base_axial_spring(self) -> pd.DataFrame:
        """Table with Q-z (base axial) spring computed for the given Model with Q-value [kN] and z-value [m].

        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with Q-z spring.
        """
        if self.soil is None:
            return None
        else:
            spring_dim = self._qz_spring.shape[-1]

            column_values_spring = [f"VAL {i}" for i in range(spring_dim)]

            df = pd.DataFrame(
                data={
                    "Node no.": [self.element_number + 1] * 2,
                    "Elevation [m]": [self.pile.bottom_elevation] * 2,
                }
            )
            df["type"] = ["q", "z"]
            df[column_values_spring] = self._qz_spring.reshape(2, spring_dim)

            return df

    def get_base_rotational_spring(self) -> pd.DataFrame:
        """Table with Mb (base moment) spring computed for the given Model with M-value [kNn] and t-value [radians].


        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with Mb spring, i.e.
        """
        if self.soil is None:
            return None
        else:
            spring_dim = self._Mb_spring.shape[-1]

            column_values_spring = [f"VAL {i}" for i in range(spring_dim)]

            df = pd.DataFrame(
                data={
                    "Node no.": [self.element_number + 1] * 2,
                    "Elevation [m]": [self.pile.bottom_elevation] * 2,
                }
            )
            df["type"] = ["Mb", "t"]
            df[column_values_spring] = self._Mb_spring.reshape(2, spring_dim)

            return df

    def plot(self, ax=None):
        """Create a plot of the model with the mesh and boundary conditions.

        Parameters
        ----------
        ax : axis handle from matplotlib figure, optional
            if None, a new axis handle is created

        Examples
        --------

        *Plot without SoilProfile fed to the model:*

        .. plot::
            :context: close-figs
            :include-source: False

            from openpile.construct import Model

            # save the figure generated by Model.plot()
            m = Model(name="<Model name>", pile=p)
            m.set_pointload(elevation=0, Py=-500)
            m.set_support(elevation=-40, Tz=True, Ty=True)
            m.set_support(elevation=-10, Ty=True)
            m.plot()

        *Plot with SoilProfile fed to the model:*

        .. plot::
            :context: close-figs
            :include-source: False

            m = Model(name="<Model name>", pile=p, soil=sp)
            m.set_pointload(elevation=0, Py=500)
            m.plot()
        """
        graphics.connectivity_plot(self, ax=ax)

    def solve(self):
        """
        Solves the boundary conditions by calling either:
         - :py:func:`openpile.analyze.beam` if no SoilProfile is provided
         - :py:func:`openpile.analyze.winkler` if a SoilProfile is provided

        Returns
        -------
        WinklerResult
            objects that stores results of the analysis.
        """
        from openpile.winkler import beam, winkler

        return beam(self) if self.soil is None else winkler(self)

    @classmethod
    def create(
        cls,
        name: str,
        pile: Pile,
        soil: Optional[SoilProfile] = None,
        element_type: Literal["Timoshenko", "EulerBernoulli"] = "Timoshenko",
        x2mesh: List[float] = Field(default_factory=list),
        coarseness: float = 0.5,
        distributed_lateral: bool = True,
        distributed_moment: bool = True,
        distributed_axial: bool = False,
        base_shear: bool = True,
        base_moment: bool = True,
        base_axial: bool = False,
    ):
        """A method to create the Model.

        Parameters
        ----------
        name : str
            Name of the model
        pile : Pile
            Pile instance to be included in the model
        soil : Optional[SoilProfile], optional
            SoilProfile instance, by default None
        element_type : str, optional
            can be of ['EulerBernoulli','Timoshenko'], by default 'Timoshenko'
        x2mesh : List[float], optional
            additional elevations to be included in the mesh, by default none
        coarseness : float, optional
            maximum distance in meters between two nodes of the mesh, by default 0.5
        distributed_lateral : bool, optional
            include distributed lateral springs, by default True
        distributed_moment : bool, optional
            include distributed moment springs, by default False
        base_shear : bool, optional
            include lateral spring at pile toe, by default False
        base_moment : bool, optional
            include moment spring at pile toe, by default False

        Returns
        -------
        openpile.construct.Model
            a Model instance with a Pile structure and optionally a SoilProfile
        """

        obj = cls(
            name=name,
            pile=pile,
            soil=soil,
            element_type=element_type,
            x2mesh=x2mesh,
            coarseness=coarseness,
            distributed_lateral=distributed_lateral,
            distributed_moment=distributed_moment,
            distributed_axial=distributed_axial,
            base_shear=base_shear,
            base_moment=base_moment,
            base_axial=base_axial,
        )

        warnings.warn(
            "\nThe method Model.create() will be removed in version 1.0.0."
            "\nPlease use the base class to create a Pile instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return obj

    def __str__(self):
        return self.element_properties.to_string()



"""
`tz_curves` module
------------------


"""

# Import libraries
from openpile.utils.misc import _fmax_api_clay, _fmax_api_sand

import math as m
import numpy as np
from numba import njit, prange
from random import random

# Kraft et al (1989) formulation aid to ease up implementation
def kraft_modification(
    fmax: float,
    D: float,
    G0: float,
    residual: float = 1.0,
    tensile_factor: float = 1.0,
    RF: float = 0.9,
    zif: float = 10.0,
    output_length: int = 15,
):
    """
    Creates the t-z curve from relevant input with the Kraft et al (1981) formulation.

    Parameters
    ----------
    fmax: float
        unit skin friction [unit: kPa]
    D: float
        Pile diameter [unit: m]
    G0: float
        small-strain stiffness [unit: kPa]
    residual: float
        residual strength after peak strength
    tensile_factor: float
        strength factor for negative values of the curve
    RF: float
        curve fitting factor as per Kraft et al. (1981), by default 0.9
    zif: float
        dimensionless zone of influence as per Kraft et al (1981) that corresponds to the radius of the zone of influence divided by the pile radius, by default 10.0
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    """

    if output_length < 15:
        output_length = 15
    if output_length % 2 == 0:
        output_length += 1

    half_length = round(output_length / 2) - 2

    # define t till tmax
    tpos = np.linspace(0, fmax, half_length)
    # define z till zmax
    zpos = tpos * 0.5 * D / G0 * np.log((zif - RF * tpos / fmax) / (1 - RF * tpos / fmax))
    # define z where t = tmax, a.k.a zmax here
    zmax = fmax * D / (2 * G0) * m.log((zif - RF) / (1 - RF))
    # define z where z=tres, which is zmax + 5mm
    zres = zmax + 0.005

    z = np.append(zpos, [zres, zres + 0.005])
    t = np.append(tpos, [residual * fmax, residual * fmax])

    return np.append(-z[-1:0:-1], z), np.append(-t[-1:0:-1] * tensile_factor, t)


# API clay backbone fct
def backbone_api_clay(
    residual: float = 0.9,
    tensile_factor: float = 1.0,
    output_length: int = 15,
):
    """
    Creates the API clay t-z curve backbone (i.e. normalized with strength and diameter) from relevant input as per [API2000]_.

    Parameters
    ----------
    D: float
        Pile width [unit: m]
    residual: float
        residual strength after peak strength, according to API-RP-2A,
        this value is between 0.7 and 0.9, default to 0.9
    tensile_factor: float
        strength factor for negative values of the curve
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]


    """
    # cannot have less than 15
    if output_length < 15:
        output_length = 15

    # piecewise function
    zlist = [0.0, 0.0016, 0.0031, 0.0057, 0.0080, 0.0100, 0.0200, 0.0300]
    tlist = [0.0, 0.3, 0.5, 0.75, 0.90, 1.00, residual, residual]

    # determine z vector
    z = np.array(zlist, dtype=np.float32)
    z = np.concatenate((-z[-1:0:-1], z))
    # define t vector
    t = np.array(tlist, dtype=np.float32)
    t = np.concatenate((-tensile_factor * t[-1:0:-1], t))

    add_values = output_length - len(z)
    add_z_values = np.zeros((add_values), dtype=np.float32)
    add_t_values = np.zeros((add_values), dtype=np.float32)

    for i in range(add_values):
        add_z_values[i] = 0.02 + random() * 0.01
        add_t_values[i] = residual

    z = np.append(z, add_z_values)
    t = np.append(t, add_t_values)

    z = np.sort(z)
    z_id_sorted = np.argsort(z)

    t = t[z_id_sorted]

    return z, t


# API clay function
def api_clay(
    sig: float,
    Su: float,
    D: float,
    alpha_limit: float = 1.0,
    residual: float = 0.9,
    tensile_factor: float = 1.0,
    output_length: int = 15,
):
    r"""
    Creates the API clay t-z curve from relevant input as per [API2000]_.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    Su : float
        Undrained shear strength [unit: kPa]
    D: float
        Pile width [unit: m]
    alpha_limit: float
        maximum value of the alpha parameter, default to 1.0
    residual: float
        residual strength after peak strength, according to API-RP-2A,
        this value is between 0.7 and 0.9, default to 0.9
    tensile_factor: float
        strength factor for negative values of the curve
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    Notes
    -----

    The backbone curve is computed via the piecewise formulation by [API2000]_ or through
    Kraft's formulation that captures small-strain characteristics of the soil [KrRK81]_ in the backbone curve.

    .. note::
        Kraft's formulation is accessed by the user by stipulating the small-strain shear
        stiffness of the soil, :math:`G_0`


    The API guidelines describe the axial soil springs in a manner that accounts for the undrained shear strength of the clay.
    The springs are characterized as follows:

    1. **Unit Skin Friction** :math:`f_s`: For clay, this is based on the undrained shear strength :math:`S_u` of the soil and a factor that accounts for the adhesion between the clay and the pile.

        .. math::

            f_s = \alpha \cdot S_u < f_{s,\text{max}}

        where:

        - :math:`\alpha` is the adhesion factor, which depends on the type of clay and the pile material.
          It typically ranges from 0.5 to 1.0 for soft clays and 0.3 to 0.6 for stiff clays.
          As per the API guidelines, this adhesion factor can be calculated as:
        - :math:`S_u` is the undrained shear strength of the clay.
        - Limit Skin Friction :math:`f_{s,\text{max}}` is the maximum unit skin friction for clay,
          which can be directly related to the undrained shear strength and the adhesion factor.
          In general, the limit skin friction is set to the undrained shear strength.

    2. A backbone curve computed via the piecewise formulation seen in [API2000]_.

    """
    # cannot have less than 15
    if output_length < 15:
        output_length = 15

    # unit skin friction
    f = _fmax_api_clay(sig, Su, alpha_limit)

    # call backbone curve
    z, t = backbone_api_clay(residual, tensile_factor, output_length)

    return z * D, t * f


@njit(cache=True)
def api_clay_kraft(
    sig: float,
    Su: float,
    D: float,
    G0: float,
    alpha_limit: float = 1.0,
    residual: float = 1.0,
    tensile_factor: float = 1.0,
    RF: float = 0.9,
    zif: float = 10.0,
    output_length: int = 15,
):
    r"""
    Creates the API clay t-z curve (see [API2000]_) with the Kraft et al (1981) formulation (see [KrRK81]_).

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    Su : float
        Undrained shear strength [unit: kPa]
    D: float
        Pile diameter [unit: m]
    G0: float
        small-strain stiffness [unit: kPa]
    alpha_limit: float
        limit of the alpha parameter, by default 1.0
    residual: float
        residual strength after peak strength, by default 1.0
    tensile_factor: float
        strength factor for negative values of the curve, by default 1.0
    RF: float
        curve fitting factor as per Kraft et al. (1981), by default 0.9
    zif: float
        dimensionless zone of influence as per Kraft et al (1981) that corresponds to the radius of the zone of influence divided by the pile radius, by default 10.0
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    See also
    --------
    :py:func:`openpile.utils.tz_curves.api_clay`

    Notes
    -----

    The backbone curve is computed via the piecewise formulation
    by [API2000]_ or through
    Kraft's formulation that captures small-strain
    characteristics of the soil [KrRK81]_ in the backbone curve.

    .. note::
        Kraft's formulation is accessed by the user by stipulating the small-strain shear
        stiffness of the soil, :math:`G_0`


    For clay, the API guidelines describe the axial soil springs in a manner that accounts for the undrained shear strength of the clay.
    The springs are characterized as follows:

    1. **Unit Skin Friction** :math:`f_s`: For clay, this is based on the undrained shear strength :math:`S_u` of the soil and a factor that accounts for the adhesion between the clay and the pile.

        .. math::

            f_s = \alpha \cdot S_u < f_{s,\text{max}}

        where:

        - :math:`\alpha` is the adhesion factor, which depends on the type of clay and the pile material.
          It typically ranges from 0.5 to 1.0 for soft clays and 0.3 to 0.6 for stiff clays.
          As per the API guidelines, this adhesion factor can be calculated as:
        - :math:`S_u` is the undrained shear strength of the clay.
        - Limit Skin Friction :math:`f_{s,\text{max}}` is the maximum unit skin friction for clay,
          which can be directly related to the undrained shear strength and the adhesion factor.
          In general, the limit skin friction is set to the undrained shear strength.

    2. A backbone curve computed via the piecewise formulation seen in [API2000]_.

    """

    return kraft_modification(
        _fmax_api_clay(sig, Su, alpha_limit),
        D,
        G0,
        residual,
        tensile_factor,
        RF,
        zif,
        output_length,
    )


# API sand function
def backbone_api_sand(
    tensile_factor: float = 1.0,
    output_length: int = 7,
):
    """
    Creates the API sand t-z curve (see [API2000]_).

    Parameters
    ----------
    tensile_factor: float
        strength factor for negative values of the curve
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 7, by default 7

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]


    """
    # cannot have less than 7
    if output_length < 15:
        output_length = 15

    # piecewise function
    zlist = [0.0, 0.0254, 0.03, 0.032, 0.034, 0.036, 0.038, 0.04]
    tlist = [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]

    # determine z vector
    z = np.array(zlist, dtype=np.float32)
    z = np.concatenate((-z[-1:0:-1], z))
    # define t vector
    t = np.array(tlist, dtype=np.float32)
    t = np.concatenate((-tensile_factor * t[-1:0:-1], t))

    add_values = output_length - len(z)
    add_z_values = np.zeros((add_values), dtype=np.float32)
    add_t_values = np.zeros((add_values), dtype=np.float32)

    for i in range(add_values):
        add_z_values[i] = 0.038 + random() * 0.01
        add_t_values[i] = 1.0

    z = np.append(z, add_z_values)
    t = np.append(t, add_t_values)

    z = np.sort(z)
    z_id_sorted = np.argsort(z)

    t = t[z_id_sorted]

    return z, t


# API sand function
def api_sand(
    sig: float,
    delta: float,
    K: float = 0.8,
    tensile_factor: float = 1.0,
    output_length: int = 7,
):
    r"""
    Creates the API sand t-z curve (see [API2000]_).

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    delta: float
        interface friction angle [unit: degrees]
    K: float
        coefficient of lateral pressure (0.8 for open-ended piles and 1.0 for cloased-ended)
    tensile_factor: float
        strength factor for negative values of the curve
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 7, by default 7

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    Notes
    -----

    The backbone curve is computed via the piecewise formulation
    by [API2000]_ or through
    Kraft's formulation that captures small-strain
    characteristics of the soil [KrRK81]_ in the backbone curve.

    .. note::
        Kraft's formulation is accessed by the user by stipulating the small-strain shear
        stiffness of the soil, :math:`G_0`

    The API guidelines provide methods to estimate the resistance offered by sandy soils along the pile.
    These springs are defined based on the following considerations:

    1. **Unit Skin Friction** :math:`f_s`: This is the frictional resistance per unit area along the pile shaft. It depends on the effective overburden pressure and the soil-pile interface properties.

        .. math::

        f_s = \sigma_v^\prime \cdot K \cdot tan(\delta) < f_{s,\text{max}}

        where:

        - \sigma_v^\prime is the effective vertical stress at the depth considered.
        - K is the coefficient of horizontal earth pressure (typically ranges from 0.4 to 1.0 for sands).
        - \delta is the angle of friction between the pile and the sand, often taken as a fraction of the soil's internal friction angle :math:`\varphi` (usually :math:`\delta = 0.7 \varphi` to :math:`\varphi`).
        - Limit Skin Friction :math:`f_{s,\text{max}}` is the maximum unit skin friction that can be mobilized. It is typically given by empirical correlations or laboratory tests. The following is assumed in OpenPile:

        .. list-table:: Correlation between interface friction angle and shaft friction
            :header-rows: 0

            * - :math:`\delta` [degrees]
              - 15
              - 20
              - 25
              - 30
              - 35
            * - :math:`f_{s,\texttt{max}}` [kPa]
              - 47.8
              - 67
              - 81.3
              - 95.7
              - 114.8

    2. A backbone curve computed via the piecewise formulation seen in [API2000]_.

    """

    # unit skin friction
    f = _fmax_api_sand(sig, delta, K)

    # call backbone curve
    z, t = backbone_api_sand(tensile_factor, output_length)

    return z, t * f


@njit(cache=True)
def api_sand_kraft(
    sig: float,
    delta: float,
    D: float,
    G0: float,
    K: float = 0.8,
    residual: float = 1.0,
    tensile_factor: float = 1.0,
    RF: float = 0.9,
    zif: float = 10.0,
    output_length: int = 15,
):
    r"""
    Creates the API sand t-z curve (see [API2000]_) with the Kraft et al (1981) formulation (see [KrRK81]_).

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    delta: float
        interface friction angle [unit: degrees]
    D: float
        Pile width [unit: m]
    G0: float
        small-strain stiffness [unit: kPa]
    K: float
        coefficient of lateral pressure (0.8 for open-ended piles and 1.0 for cloased-ended), by default 0.8
    residual: float
        residual strength after peak strength, by default 1.0
    tensile_factor: float
        strength factor for negative values of the curve, by default 1.0
    RF: float
        curve fitting factor as per Kraft et al. (1981), by default 0.9
    zif: float
        dimensionless zone of influence as per Kraft et al (1981) that corresponds to the radius of the zone of influence divided by the pile radius, by default 10.0
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    See also
    --------
    :py:func:`openpile.utils.tz_curves.api_sand`

    Notes
    -----

    The backbone curve is computed via the piecewise formulation
    by [API2000]_ or through
    Kraft's formulation that captures small-strain
    characteristics of the soil [KrRK81]_ in the backbone curve.

    .. note::
        Kraft's formulation is accessed by the user by stipulating the small-strain shear
        stiffness of the soil, :math:`G_0`

    The API guidelines provide methods to estimate the resistance offered by sandy soils along the pile.
    These springs are defined based on the following considerations:

    1. **Unit Skin Friction** :math:`f_s`: This is the frictional resistance per unit area along the pile shaft. It depends on the effective overburden pressure and the soil-pile interface properties.

        .. math::

        f_s = \sigma_v^\prime \cdot K \cdot tan(\delta) < f_{s,\text{max}}

        where:

        - \sigma_v^\prime is the effective vertical stress at the depth considered.
        - K is the coefficient of horizontal earth pressure (typically ranges from 0.4 to 1.0 for sands).
        - \delta is the angle of friction between the pile and the sand, often taken as a fraction of the soil's internal friction angle :math:`\varphi` (usually :math:`\delta = 0.7 \varphi` to :math:`\varphi`).
        - Limit Skin Friction :math:`f_{s,\text{max}}` is the maximum unit skin friction that can be mobilized. It is typically given by empirical correlations or laboratory tests. The following is assumed in OpenPile:

        .. list-table:: Correlation between interface friction angle and shaft friction
            :header-rows: 0

            * - :math:`\delta` [degrees]
              - 15
              - 20
              - 25
              - 30
              - 35
            * - :math:`f_{s,\texttt{max}}` [kPa]
              - 47.8
              - 67
              - 81.3
              - 95.7
              - 114.8

    2. A backbone curve computed via the piecewise formulation seen in [API2000]_.

    """
    return kraft_modification(
        _fmax_api_sand(sig, delta, K), D, G0, residual, tensile_factor, RF, zif, output_length
    )



# misc functions


import random
import numpy as np
import pandas as pd
import math as m

from numba import njit
import matplotlib.colors as mcolors


def from_list2x_parse_top_bottom(var):
    """provide top and bottom values of layer based on float or list inputs"""
    if isinstance(var, float) or isinstance(var, int):
        top = var
        bottom = var
    elif isinstance(var, list) and len(var) == 1:
        top = var[0]
        bottom = var[0]
    elif isinstance(var, list) and len(var) == 2:
        top = var[0]
        bottom = var[1]
    else:
        print("Soil Layer variable is not a float nor a list")
        raise TypeError

    return top, bottom


def get_value_at_current_depth(X, depth_from_top_of_layer, layer_height, depth_from_ground):
    if isinstance(X, type(lambda x: 2)):
        return X(depth_from_ground)
    else:
        xtop, xbot = from_list2x_parse_top_bottom(X)
        return xtop + (xbot - xtop) * depth_from_top_of_layer / layer_height


def var_to_str(var):
    if isinstance(var, float) or isinstance(var, int):
        var_print = var
    elif isinstance(var, list):
        var_print = "-".join(str(v) for v in var)
    else:
        raise ValueError("not a float nor list")
    return var_print


def generate_color_string(kind=None):
    if kind is None:
        colors = list(mcolors.CSS4_COLORS.values())
    elif kind == "earth":
        colors = [
            "#B4A390",
            "#927E75",
            "#796363",
            "#99A885",
            "#D4DBBA",
            "#FDEDCF",
            "#EEEAC3",
            "#F5D498",
            "#ECB992",
            "#DDA175",
            "#AB7B5E",
            "#8F6854",
        ]
    return colors[random.randint(0, len(colors) - 1)]


def repeat_inner(arr):
    arr = arr.reshape(-1, 1)

    arr_inner = arr[1:-1]
    arr_inner = np.tile(arr_inner, (2)).reshape(-1)

    return np.hstack([arr[0], arr_inner, arr[-1]])


def get_distributed_soil_springs(
    springs: np.ndarray, elevations: np.ndarray, kind: str
) -> pd.DataFrame:
    """
    Returns soil springs created for the given model in one DataFrame.

    Unit of the springs are [kN/m]/[kNm/m] for resistance and [m]/[rad] for displacement/rotation.

    Parameters
    ----------
    springs : ndarray dim[nelem,2,2,spring_dim]
        Springs at top and bottom of element
    elevations : ndarray
        self.nodes_coordinates["z [m]"].values
    kind : str
        type of spring to extract. one of ["p-y", "m-t", "Hb-y", "Mb-t", "t-z"]

    Returns
    -------
    pd.DataFrame
        Soil springs
    """

    # get rid of the dimension dependent on the p-value, we keep max m-t spring
    if kind == "m-t":
        springs = springs.max(axis=3)

    spring_dim = springs.shape[-1]
    nelem = springs.shape[0]
    nnode = len(elevations)

    column_values_spring = [f"VAL {i}" for i in range(spring_dim)]

    id = np.repeat(np.arange(nelem + 1), 2)
    x = np.repeat(elevations, 2)

    influence = np.abs(np.gradient(elevations))
    influence[0] = influence[0] / 2
    influence[-1] = influence[-1] / 2

    springs[:, 0, 0, :] = springs[:, 0, 0, :] * influence[:-1].reshape(-1, 1)
    springs[:, 1, 0, :] = springs[:, 1, 0, :] * influence[1:].reshape(-1, 1)

    reduced_springs = np.zeros((nnode * 2, spring_dim))

    # first spring resistance and disp values
    reduced_springs[0, :] = springs[0, 0, 0, :]
    reduced_springs[1, :] = springs[0, 0, 1, :]
    # last spring resistance and disp values
    reduced_springs[-2, :] = springs[-1, 1, 0, :]
    reduced_springs[-1, :] = springs[-1, 1, 1, :]
    # calculation of weighted springs when node based
    j = 0
    for i in range(2, nelem * 2 - 1, 2):
        j += 1
        reduced_springs[i, :] = (
            springs[j - 1, 1, 0, :] * influence[j - 1] + springs[j, 0, 0, :] * influence[j]
        ) / (influence[j - 1] + influence[j])
        reduced_springs[i + 1, :] = (
            springs[j - 1, 1, 1, :] * influence[j - 1] + springs[j, 0, 1, :] * influence[j]
        ) / (influence[j - 1] + influence[j])

    df = pd.DataFrame(
        data={
            "Node no.": id,
            "Elevation [m]": x,
        }
    )

    df["type"] = kind.split("-") * len(elevations)
    df[column_values_spring] = reduced_springs

    return df


def get_lumped_soil_springs(springs: np.ndarray, elevations: np.ndarray, kind: str) -> pd.DataFrame:
    """
    Returns soil springs created for the given model in one DataFrame.

    Unit of the springs are [kN]/[kNm] for resistance and [m]/[rad] for displacement/rotation.

    Parameters
    ----------
    springs : ndarray dim[nelem,2,2,spring_dim]
        Springs at top and bottom of element
    elevations : ndarray
        self.nodes_coordinates["z [m]"].values
    kind : str
        type of spring to extract. one of ["p-y", "m-t", "Hb-y", "Mb-t", "t-z"]

    Returns
    -------
    pd.DataFrame
        Soil springs
    """

    # get rid of the dimension dependent on the p-value, we keep max m-t spring
    if kind == "m-t":
        springs = springs.max(axis=3)

    spring_dim = springs.shape[-1]
    nelem = springs.shape[0]
    nnode = len(elevations)

    column_values_spring = [f"VAL {i}" for i in range(spring_dim)]

    id = np.repeat(np.arange(nelem + 1), 2)
    x = np.repeat(elevations, 2)

    influence = np.abs(np.gradient(elevations))
    influence[0] = influence[0] / 2
    influence[-1] = influence[-1] / 2

    springs[:, 0, 0, :] = springs[:, 0, 0, :] * influence[:-1].reshape(-1, 1)
    springs[:, 1, 0, :] = springs[:, 1, 0, :] * influence[1:].reshape(-1, 1)

    reduced_springs = np.zeros((nnode * 2, spring_dim))

    # first spring resistance and disp values
    reduced_springs[0, :] = springs[0, 0, 0, :]
    reduced_springs[1, :] = springs[0, 0, 1, :]
    # last spring resistance and disp values
    reduced_springs[-2, :] = springs[-1, 1, 0, :]
    reduced_springs[-1, :] = springs[-1, 1, 1, :]
    # calculation of weighted springs when node based
    j = 0
    for i in range(2, nelem * 2 - 1, 2):
        j += 1
        reduced_springs[i, :] = (
            springs[j - 1, 1, 0, :] * influence[j - 1] + springs[j, 0, 0, :] * influence[j]
        )
        reduced_springs[i + 1, :] = (
            springs[j - 1, 1, 1, :] * influence[j - 1] + springs[j, 0, 1, :] * influence[j]
        )

    df = pd.DataFrame(
        data={
            "Node no.": id,
            "Elevation [m]": x,
        }
    )

    df["type"] = kind.split("-") * len(elevations)
    df[column_values_spring] = reduced_springs

    return df


@njit(cache=True)
def conic(
    x_u: float,
    n: float,
    k: float,
    y_u: float,
    output_length: int,
):
    # if k is less than y_u/x_u, k overwritten to equate y_u/x_u
    k = max(y_u / x_u, k)

    # Create x vector with 10% extension
    x = np.array([0, 0.02, 0.05, 0.1]).astype(np.float32) * x_u
    x = np.append(x, np.linspace(0.2 * x_u, x_u, output_length - 5).astype(np.float32))
    x = np.append(x, 1.1 * x_u)

    a = 1 - 2 * n

    y = np.zeros((len(x)), dtype=np.float32)

    for i in range(len(x)):
        if abs(x[i] - x_u) < 1e-2:
            y[i] = y_u
        elif x[i] < x_u:
            b = 2 * n * x[i] / x_u - (1 - n) * (1 + x[i] * k / y_u)
            c = x[i] * (k / y_u) * (1 - n) - n * (x[i] ** 2 / x_u**2)

            y[i] = y_u * 2 * c / (-b + (b**2 - 4 * a * c) ** 0.5)
        else:
            y[i] = y_u

    return x, y



import pytest
from openpile.construct import Pile, SoilProfile, Layer, Model
from openpile.soilmodels import API_sand
from openpile.winkler import winkler
from openpile.utils.hooks import durkhop

import math as m


@pytest.fixture
def create_pile():
    return Pile.create_tubular(
        name="<pile name>", top_elevation=0, bottom_elevation=-40, diameter=7, wt=0.050
    )


@pytest.fixture
def create_cyclic_soilprofile():
    # Create a 40m deep offshore Soil Profile with a 15m water column
    sp = SoilProfile(
        name="Offshore Soil Profile",
        top_elevation=0,
        water_line=15,
        layers=[
            Layer(
                name="medium dense sand",
                top=0,
                bottom=-40,
                weight=18,
                lateral_model=API_sand(phi=33, kind="cyclic"),
            ),
        ],
    )
    return sp


@pytest.fixture
def create_static_soilprofile():
    # Create a 40m deep offshore Soil Profile with a 15m water column
    sp = SoilProfile(
        name="Offshore Soil Profile",
        top_elevation=0,
        water_line=15,
        layers=[
            Layer(
                name="medium dense sand",
                top=0,
                bottom=-40,
                weight=18,
                lateral_model=API_sand(phi=33, kind="static"),
            ),
        ],
    )
    return sp


@pytest.fixture
def create_duhrkop_soilprofile():
    # Create a 40m deep offshore Soil Profile with a 15m water column
    sp = SoilProfile(
        name="Offshore Soil Profile",
        top_elevation=0,
        water_line=15,
        layers=[
            Layer(
                name="medium dense sand",
                top=0,
                bottom=-40,
                weight=18,
                lateral_model=API_sand(phi=33, kind="cyclic", p_multiplier=durkhop(D=7, ra=0.3)),
            ),
        ],
    )
    return sp


def test_APIcylic_equals_durkhop_ra_03(
    create_pile, create_cyclic_soilprofile, create_duhrkop_soilprofile
):

    # Create Model 1
    M1 = Model(name="<model name>", pile=create_pile, soil=create_cyclic_soilprofile)
    # Apply bottom fixity along x-axis
    M1.set_support(elevation=-40, Tz=True)
    # Apply axial and lateral loads
    M1.set_pointload(elevation=0, Mx=-300e3, Py=10e3)
    R1 = winkler(M1)

    # Create Model 2
    M2 = Model(name="<model name>", pile=create_pile, soil=create_duhrkop_soilprofile)
    # Apply bottom fixity along x-axis
    M2.set_support(elevation=-40, Tz=True)
    # Apply axial and lateral loads
    M2.set_pointload(elevation=0, Mx=-300e3, Py=10e3)
    R2 = winkler(M2)

    assert m.isclose(
        R1.details()["Max. deflection [m]"], R2.details()["Max. deflection [m]"], rel_tol=0.01
    )


def test_pisa_custom_sand_py():

    from openpile.utils.py_curves import custom_pisa_sand, dunkirk_sand
    from openpile.utils.hooks import PISA_depth_variation

    # plot original dunkirk sand curve
    params = {"sig": 50, "X": 5, "Dr": 75, "G0": 50e3, "D": 6, "L": 20}
    y, p = dunkirk_sand(**params)
    # load PISA dunkirk sand depth variation functions
    funcs = PISA_depth_variation.dunkirk_sand_py_pisa_norm_param(D=6, L=20, Dr=75)
    # plot same curve with custom pisa sand curve
    params = {
        "sig": 50,
        "G0": 50e3,
        "D": 6,
        "X_ult": funcs["py_X"](5),
        "Y_ult": funcs["py_Y"](5),
        "n": funcs["py_n"](5),
        "k": funcs["py_k"](5),
    }
    y_custom, p_custom = custom_pisa_sand(**params)

    for i in range(len(y)):
        assert m.isclose(y[i], y_custom[i])
        assert m.isclose(p[i], p_custom[i])


def test_pisa_custom_sand_mt():

    from openpile.utils.mt_curves import custom_pisa_sand, dunkirk_sand
    from openpile.utils.hooks import PISA_depth_variation

    # plot original dunkirk sand curve
    params = {"sig": 50, "X": 5, "Dr": 75, "G0": 50e3, "D": 6, "L": 20, "p": 100}
    y, p = dunkirk_sand(**params)
    # load PISA dunkirk sand depth variation functions
    funcs = PISA_depth_variation.dunkirk_sand_mt_pisa_norm_param(L=20, Dr=75)
    # plot same curve with custom pisa sand curve
    params = {
        "sig": 50,
        "G0": 50e3,
        "D": 6,
        "p": 100,
        "X_ult": funcs["mt_X"](5),
        "Y_ult": funcs["mt_Y"](5),
        "n": funcs["mt_n"](5),
        "k": funcs["mt_k"](5),
    }
    y_custom, p_custom = custom_pisa_sand(**params)

    for i in range(len(y)):
        assert m.isclose(y[i], y_custom[i])
        assert m.isclose(p[i], p_custom[i])


def test_pisa_custom_clay_py():

    from openpile.utils.py_curves import custom_pisa_clay, cowden_clay
    from openpile.utils.hooks import PISA_depth_variation

    # plot original dunkirk sand curve
    params = {"X": 5, "Su": 75, "G0": 50e3, "D": 6}
    y, p = cowden_clay(**params)
    # load PISA dunkirk sand depth variation functions
    funcs = PISA_depth_variation.cowden_clay_py_pisa_norm_param(D=6)
    # plot same curve with custom pisa sand curve
    params = {
        "Su": 75,
        "G0": 50e3,
        "D": 6,
        "X_ult": funcs["py_X"](5),
        "Y_ult": funcs["py_Y"](5),
        "n": funcs["py_n"](5),
        "k": funcs["py_k"](5),
    }
    y_custom, p_custom = custom_pisa_clay(**params)

    for i in range(len(y)):
        assert m.isclose(y[i], y_custom[i])
        assert m.isclose(p[i], p_custom[i])


def test_pisa_custom_clay_mt():

    from openpile.utils.mt_curves import custom_pisa_clay, cowden_clay
    from openpile.utils.hooks import PISA_depth_variation

    # plot original dunkirk sand curve
    params = {"X": 5, "Su": 75, "G0": 50e3, "D": 6}
    y, p = cowden_clay(**params)
    # load PISA dunkirk sand depth variation functions
    funcs = PISA_depth_variation.cowden_clay_mt_pisa_norm_param(D=6)
    # plot same curve with custom pisa sand curve
    params = {
        "Su": 75,
        "G0": 50e3,
        "D": 6,
        "X_ult": funcs["mt_X"](5),
        "Y_ult": funcs["mt_Y"](5),
        "n": funcs["mt_n"](5),
        "k": funcs["mt_k"](5),
    }
    y_custom, p_custom = custom_pisa_clay(**params)

    for i in range(len(y)):
        assert m.isclose(y[i], y_custom[i])
        assert m.isclose(p[i], p_custom[i])



"""
---------------
`Kernel` module
---------------

#TODO

"""
# general utilities for openfile

from copy import deepcopy
import numpy as np
import pandas as pd
from numba import njit, prange, f4, f8, b1, char
import numba as nb
from typing_extensions import Literal

from openpile.construct import CircularPileSection
from openpile.core.validation import UserInputError
from openpile.core import misc
from openpile.core._model_build import parameter2elements


@njit(f8[:](f8[:, :], f8[:]), cache=True)
def jit_solve(A, b):
    return np.linalg.solve(A, b)


@njit(parallel=True, cache=True)
def jit_eigh(A):
    return np.linalg.eigh(A)


@njit(parallel=True, cache=True)
def reverse_indices(A, B):
    return np.array([x for x in A if x not in B])


@njit(f8[:](f8[:]), cache=True)
def double_inner_njit(in_arr):
    # d = misc.repeat_inner(u)
    out_arr = np.zeros(((len(in_arr) - 1) * 2), dtype=np.float64)
    i = 0
    while i < len(out_arr) + 1:
        if i == 0:
            out_arr[i] = in_arr[0]
            i += 1
        elif i == len(out_arr):
            out_arr[i] = in_arr[-1]
            i += 1
        elif i == len(out_arr) - 1:
            out_arr[i] = in_arr[-2]
            i += 1
        elif i % 2 != 0:
            x = int((i + 1) / 2)
            out_arr[i] = in_arr[x]
            out_arr[i + 1] = in_arr[x]
            i += 2
        elif i % 2 == 0:
            x = int(i / 2)
            out_arr[i] = in_arr[x]
            out_arr[i + 1] = in_arr[x]
            i += 2

    return out_arr


@njit(parallel=True, cache=True)
def numba_ix(arr, rows, cols):
    """
    Numba compatible implementation of arr[np.ix_(rows, cols)] for 2D arrays.
    :param arr: 2D array to be indexed
    :param rows: Row indices
    :param cols: Column indices
    :return: 2D array with the given rows and columns of the input array
    """
    one_d_index = np.zeros(len(rows) * len(cols), dtype=np.int32)
    for i, r in enumerate(rows):
        start = i * len(cols)
        one_d_index[start : start + len(cols)] = cols + arr.shape[1] * r

    arr_1d = arr.reshape((arr.shape[0] * arr.shape[1], 1))
    slice_1d = np.take(arr_1d, one_d_index)
    return slice_1d.reshape((len(rows), len(cols)))


def global_dof_vector_to_consistent_stacked_array(dof_vector, dim):
    arr = dof_vector.reshape((-1, 3))
    arr_inner = np.tile(arr[1:-1], 2)
    arr = np.vstack([arr[0, :], arr_inner.reshape((-1, 3)), arr[-1, :]])
    arr = arr.reshape(-1, dim, 1)

    return arr


def solve_equations(K, F, U, restraints):
    r"""function that solves the system of equations

    The function uses numba to speed up the computational time.

    The system of equations to solve is the following :eq:`solve_eq_3dof_example` assuming a trivial 3 degree of freedom system:
    
    .. math::
        :label: solve_eq_3dof_example
    
        \begin{pmatrix}
        U_1\\ 
        U_2\\ 
        U_3
        \end{pmatrix} = 
        \begin{bmatrix}
        K_{11} & K_{12} & K_{13} \\ 
        K_{21} & K_{22} & K_{23}\\ 
        K_{31} & K_{23} & K_{33}
        \end{bmatrix} \cdot \begin{pmatrix}
        F_1 \\ 
        F_2 \\ 
        F_3
        \end{pmatrix}
        \\
        \\

    Parameters
    ----------
    K : float numpy array of dim (ndof, ndof)
        Global stiffness matrix with all dofs in unit:1/kPa.
    F : float numpy array of dim (ndof, 1)
        External force vetor (can also be denoted as load vector) in unit:kN.
    restraints: boolean numpy array of dim (ndof, 1)
        External vector of restrained dof.
        
    Returns
    -------
    U : float numpy array of dim (ndof, 1)
        Global displacement vector in unit:m.
    Q : float numpy array of dim (ndof, 1)
        Global reaction force vector in unit:kN
    
    """

    if restraints.any():
        prescribed_dof_true = np.where(restraints)[0]
        prescribed_dof_false = np.where(~restraints)[0]

        prescribed_disp = U

        Fred = F[prescribed_dof_false] - numba_ix(K, prescribed_dof_false, prescribed_dof_true).dot(
            prescribed_disp[prescribed_dof_true]
        )
        Kred = numba_ix(K, prescribed_dof_false, prescribed_dof_false)

        U[prescribed_dof_false] = jit_solve(Kred, Fred)
    else:
        U = jit_solve(K, F)

    Q = K.dot(U) - F

    return U, Q


def mesh_to_element_length(model) -> np.ndarray:

    # initialize element properties
    elem_prop = model.element_properties

    # elememt coordinates along z and y-axes
    ez = np.array(
        [
            elem_prop["z_top [m]"].to_numpy(dtype=float),
            elem_prop["z_bottom [m]"].to_numpy(dtype=float),
        ]
    )
    ey = np.array(
        [
            elem_prop["y_top [m]"].to_numpy(dtype=float),
            elem_prop["y_bottom [m]"].to_numpy(dtype=float),
        ]
    )
    # length calcuated via pythagorus theorem
    L = np.sqrt(((ez[0] - ez[1]) ** 2 + (ey[0] - ey[1]) ** 2)).reshape(-1, 1, 1)

    return L


def elem_mechanical_stiffness_matrix(model):
    """creates pile element stiffness matrix based on model info and element number

    Parameters
    ----------
    model : openpile class object `openpile.construct.Model`
        includes information on soil/structure, elements, nodes and other model-related data.

    Returns
    -------
    k: numpy array (3d)
        element stiffness matrix of all elememts related to the mechanical properties of the structure

    Raises
    ------
    ValueError
        Model.element.type only accepts 'EB' type (for Euler-Bernoulli) or 'T' type (for Timoshenko)
    ValueError
        Timoshenko beams cannot be used yet for non-circular pile types
    ValueError
        ndof per node can be either 2 or 3

    """

    # calculate length vector
    L = mesh_to_element_length(model)
    # elastic properties
    nu = model.pile.material.poisson
    E = model.pile.material.young_modulus * np.ones(L.shape).reshape((-1, 1, 1))
    G = E / (2 + 2 * nu)

    # initialize element properties
    elem_prop = model.element_properties

    # cross-section properties
    I = parameter2elements(
        model.pile.sections,
        lambda x: x.second_moment_of_area,
        elem_prop["z_top [m]"].values,
        elem_prop["z_bottom [m]"].values,
    ).reshape((-1, 1, 1))
    A = parameter2elements(
        model.pile.sections,
        lambda x: x.area,
        elem_prop["z_top [m]"].values,
        elem_prop["z_bottom [m]"].values,
    ).reshape((-1, 1, 1))

    # calculate shear component in stiffness matrix (if Timorshenko)
    if model.element_type == "EulerBernoulli":
        kappa = 0
    elif model.element_type == "Timoshenko":
        if all([isinstance(section, CircularPileSection) for section in model.pile.sections]):
            # pile diameter (called width)
            d = parameter2elements(
                model.pile.sections,
                lambda x: x.width,
                elem_prop["z_top [m]"].values,
                elem_prop["z_bottom [m]"].values,
            ).reshape((-1, 1, 1))
            # wall thickness
            wt = parameter2elements(
                model.pile.sections,
                lambda x: x.thickness,
                elem_prop["z_top [m]"].values,
                elem_prop["z_bottom [m]"].values,
            ).reshape((-1, 1, 1))

            a = 0.5 * d
            b = 0.5 * (d - 2 * wt)
            nom = 6 * (a**2 + b**2) ** 2 * (1 + nu) ** 2
            denom = (
                7 * a**4
                + 34 * a**2 * b**2
                + 7 * b**4
                + nu * (12 * a**4 + 48 * a**2 * b**2 + 12 * b**4)
                + nu**2 * (4 * a**4 + 16 * a**2 * b**2 + 4 * b**4)
            )
            kappa = nom / denom
        else:
            raise ValueError("Timoshenko beams cannot be used yet for non-circular pile types")

    phi = 12 * E * I * kappa / (A * G * L**2)
    X = A * E / L
    Y1 = (12 * E * I) / ((1 + phi) * L**3)
    Y2 = (6 * E * I) / ((1 + phi) * L**2)
    Y3 = ((4 + phi) * E * I) / ((1 + phi) * L)
    Y4 = ((2 - phi) * E * I) / ((1 + phi) * L)
    N = np.zeros(shape=X.shape)

    k = np.block(
        [
            [X, N, N, -X, N, N],
            [N, Y1, Y2, N, -Y1, Y2],
            [N, Y2, Y3, N, -Y2, Y4],
            [-X, N, N, X, N, N],
            [N, -Y1, -Y2, N, Y1, -Y2],
            [N, Y2, Y4, N, -Y2, Y3],
        ]
    )

    return k


def elem_py_stiffness_matrix(model, u, kind):
    """creates soil element stiffness matrix based on model info and element number.

    The soil stiffness matrix assumes the soil stiffness to vary lineraly along the elements.

    #TODO: proof here

    Parameters
    ----------
    model : openpile class object `openpile.construct.Model`
        includes information on soil/structure, elements, nodes and other model-related data.
    u: np.ndarray
        Global displacement vector
    kind: str
        "initial", "secant" or "tangent"

    Returns
    -------
    k: numpy array (3d)
        soil consistent stiffness matrix of all elememts related to the p-y soil springs' stiffness

    Raises
    ------
    ValueError
        ndof per node can be either 2 or 3

    """

    # calculate length vector
    L = mesh_to_element_length(model)

    # calculate the spring stiffness
    ksoil = calculate_py_springs_stiffness(u=u[1::3], springs=model._py_springs, kind=kind)

    N = 0 * L
    A = 2 * L / 7
    B = L**2 / 28
    C = L**3 / 168
    D = 9 * L / 140
    E = L**2 / 70
    F = 3 * L / 35
    G = -(L**2) / 60
    H = -(L**3) / 280
    I = L**3 / 280

    ktop = (
        np.block(
            [
                [N, N, N, N, N, N],
                [N, A, B, N, D, G],
                [N, B, C, N, E, H],
                [N, N, N, N, N, N],
                [N, D, E, N, F, G],
                [N, G, H, N, G, I],
            ]
        )
        * ksoil[:, 0]
    )

    kbottom = (
        np.block(
            [
                [N, N, N, N, N, N],
                [N, F, -G, N, D, -E],
                [N, -G, -H, N, -G, -I],
                [N, N, N, N, N, N],
                [N, D, -G, N, A, -B],
                [N, -E, -I, N, -B, C],
            ]
        )
        * ksoil[:, 1]
    )

    return ktop + kbottom


def elem_tz_stiffness_matrix(model, u, kind):
    """creates soil element stiffness matrix based on model info and element number.

    The soil stiffness matrix assumes the soil stiffness to vary lineraly along the elements.

    #TODO: proof here

    Parameters
    ----------
    model : openpile class object `openpile.construct.Model`
        includes information on soil/structure, elements, nodes and other model-related data.
    u: np.ndarray
        Global displacement vector
    kind: str
        "initial", "secant" or "tangent"

    Returns
    -------
    k: numpy array (3d)
        soil consistent stiffness matrix of all elememts related to the t-z soil springs' stiffness

    Raises
    ------
    ValueError
        ndof per node can be either 2 or 3

    """

    # calculate length vector
    L = mesh_to_element_length(model)

    # calculate the spring stiffness
    ksoil = calculate_tz_springs_stiffness(u=u[::3], springs=model._tz_springs, kind=kind)

    N = 0 * L
    A = L / 4
    B = L / 12

    ktop = (
        np.block(
            [
                [A, N, N, B, N, N],
                [N, N, N, N, N, N],
                [N, N, N, N, N, N],
                [B, N, N, B, N, N],
                [N, N, N, N, N, N],
                [N, N, N, N, N, N],
            ]
        )
        * ksoil[:, 0]
    )

    kbottom = (
        np.block(
            [
                [B, N, N, B, N, N],
                [N, N, N, N, N, N],
                [N, N, N, N, N, N],
                [B, N, N, A, N, N],
                [N, N, N, N, N, N],
                [N, N, N, N, N, N],
            ]
        )
        * ksoil[:, 1]
    )

    return ktop + kbottom


def elem_mt_stiffness_matrix(model, u, kind):
    """creates soil element stiffness matrix based on model info and element number.

    The soil stiffness matrix assumes the soil stiffness to vary lineraly along the elements.

    #TODO: proof here

    Parameters
    ----------
    model : openpile class object `openpile.construct.Model`
        includes information on soil/structure, elements, nodes and other model-related data.
    u: np.ndarray
        Global displacement vector
    kind: str
        "initial", "secant" or "tangent"

    Returns
    -------
    k: numpy array (3d)
        soil consistent stiffness matrix of all elememts related to the p-y soil springs' stiffness

    Raises
    ------
    ValueError
        ndof per node can be either 2 or 3

    """

    # calculate length vector
    L = mesh_to_element_length(model)

    # calculate the py spring stiffness
    kspy = calculate_py_springs_stiffness(u=u[1::3], springs=model._py_springs, kind="secant")
    upy = double_inner_njit(u[1::3]).reshape(-1, 2, 1, 1)
    p_mobilised = kspy * upy

    # calculate the spring stiffness
    ksoil = calculate_mt_springs_stiffness(
        u=u[2::3],
        mt_springs=model._mt_springs,
        py_springs=model._py_springs,
        p_mobilised=p_mobilised,
        kind=kind,
    )

    # initialize element properties
    elem_prop = model.element_properties

    # elastic properties
    nu = model.pile.material.poisson
    E = model.pile.material.young_modulus * np.ones(L.shape).reshape((-1, 1, 1))
    G = E / (2 + 2 * nu)
    # cross-section properties
    I = parameter2elements(
        model.pile.sections,
        lambda x: x.second_moment_of_area,
        elem_prop["z_top [m]"].values,
        elem_prop["z_bottom [m]"].values,
    ).reshape((-1, 1, 1))
    A = parameter2elements(
        model.pile.sections,
        lambda x: x.area,
        elem_prop["z_top [m]"].values,
        elem_prop["z_bottom [m]"].values,
    ).reshape((-1, 1, 1))

    # calculate shear component in stiffness matrix (if Timorshenko)
    if model.element_type == "EulerBernoulli":
        N = 0 * L
        A = 6 / (5 * L)
        B = N + 1 / 10
        C = 2 * L / 15
        D = -L / 30
    elif model.element_type == "Timoshenko":
        if all([isinstance(section, CircularPileSection) for section in model.pile.sections]):
            # pile diameter (called width)
            d = parameter2elements(
                model.pile.sections,
                lambda x: x.width,
                elem_prop["z_top [m]"].values,
                elem_prop["z_bottom [m]"].values,
            ).reshape((-1, 1, 1))
            # wall thickness
            wt = parameter2elements(
                model.pile.sections,
                lambda x: x.thickness,
                elem_prop["z_top [m]"].values,
                elem_prop["z_bottom [m]"].values,
            ).reshape((-1, 1, 1))

            a = 0.5 * d
            b = 0.5 * (d - 2 * wt)
            nom = 6 * (a**2 + b**2) ** 2 * (1 + nu) ** 2
            denom = (
                7 * a**4
                + 34 * a**2 * b**2
                + 7 * b**4
                + nu * (12 * a**4 + 48 * a**2 * b**2 + 12 * b**4)
                + nu**2 * (4 * a**4 + 16 * a**2 * b**2 + 4 * b**4)
            )
            kappa = nom / denom

            omega = E * I * kappa / (A * G * L**2)

        else:
            raise ValueError("Timoshenko beams cannot be used yet for non-circular pile types")

        phi = (12 * omega + 1) ** 2

        N = 0 * L
        A = 6 * (120 * omega**2 + 20 * omega + 1) / ((5 * L) * phi) + 12 * I / (A * L**3 * phi)
        B = 1 / (10 * phi) + 6 * I / (A * L**2 * phi)
        C = (2 * L * (90 * omega**2 + 15 * omega + 1)) / (15 * phi) + 4 * I * (
            36 * omega**2 + 6 * omega + 1
        ) / (A * L * phi)
        D = -L * (360 * omega**2 + 60 * omega + 1) / (30 * phi) - 2 * I * (
            72 * omega**2 + 12 * omega - 1
        ) / (A * L * phi)

    else:
        raise ValueError(
            "Model.element.type only accepts 'EB' type (for Euler-Bernoulli) of 'T' type (for Timoshenko)"
        )

    km = np.block(
        [
            [N, N, N, N, N, N],
            [N, A, B, N, -A, B],
            [N, B, C, N, -B, D],
            [N, N, N, N, N, N],
            [N, -A, -B, N, A, -B],
            [N, B, D, N, -B, C],
        ]
    ) * (0.5 * ksoil[:, 0] + 0.5 * ksoil[:, 1])

    return km


def elem_p_delta_stiffness_matrix(model, u):
    """creates stress stiffness matrix based on axial stress in pile.

    #TODO: proof here

    Parameters
    ----------
    model : openpile class object `openpile.construct.Model`
        includes information on soil/structure, elements, nodes and other model-related data.
    u: np.ndarray
        Global displacement vector

    Returns
    -------
    k: numpy array (3d)
        stress stiffness matrix of all elememts related to axial stress

    Raises
    ------
    ValueError
        ndof per node can be either 2 or 3

    """

    # calculate length vector
    L = mesh_to_element_length(model)

    # internal forces
    f = pile_internal_forces(model, u)
    P = f[::6].reshape(-1, 1, 1)

    # initialize element properties
    elem_prop = model.element_properties

    # elastic properties
    nu = model.pile.material.poisson
    E = model.pile.material.young_modulus * np.ones(L.shape).reshape((-1, 1, 1))
    G = E / (2 + 2 * nu)
    # cross-section properties
    I = parameter2elements(
        model.pile.sections,
        lambda x: x.second_moment_of_area,
        elem_prop["z_top [m]"].values,
        elem_prop["z_bottom [m]"].values,
    ).reshape((-1, 1, 1))
    A = parameter2elements(
        model.pile.sections,
        lambda x: x.area,
        elem_prop["z_top [m]"].values,
        elem_prop["z_bottom [m]"].values,
    ).reshape((-1, 1, 1))

    # calculate shear component in stiffness matrix (if Timorshenko)
    if model.element_type == "EulerBernoulli":
        N = 0 * L
        N1 = 0 * L
        N2 = 0 * L
        A = 6 / (5 * L)
        B = N + 1 / 10
        C = 2 * L / 15
        D = -L / 30
    elif model.element_type == "Timoshenko":
        if all([isinstance(section, CircularPileSection) for section in model.pile.sections]):
            # pile diameter (called width)
            d = parameter2elements(
                model.pile.sections,
                lambda x: x.width,
                elem_prop["z_top [m]"].values,
                elem_prop["z_bottom [m]"].values,
            ).reshape((-1, 1, 1))
            # wall thickness
            wt = parameter2elements(
                model.pile.sections,
                lambda x: x.thickness,
                elem_prop["z_top [m]"].values,
                elem_prop["z_bottom [m]"].values,
            ).reshape((-1, 1, 1))

            a = 0.5 * d
            b = 0.5 * (d - 2 * wt)
            nom = 6 * (a**2 + b**2) ** 2 * (1 + nu) ** 2
            denom = (
                7 * a**4
                + 34 * a**2 * b**2
                + 7 * b**4
                + nu * (12 * a**4 + 48 * a**2 * b**2 + 12 * b**4)
                + nu**2 * (4 * a**4 + 16 * a**2 * b**2 + 4 * b**4)
            )
            kappa = nom / denom

            omega = E * I * kappa / (A * G * L**2)

        else:
            raise ValueError("Timoshenko beams cannot be used yet for non-circular pile types")

        phi = (12 * omega + 1) ** 2

        N = 0 * L
        N1 = 0 / L
        N2 = 0 / L
        A = 6 * (120 * omega**2 + 20 * omega + 1) / ((5 * L) * phi) + 12 * I / (A * L**3 * phi)
        B = 1 / (10 * phi) + 6 * I / (A * L**2 * phi)
        C = (2 * L * (90 * omega**2 + 15 * omega + 1)) / (15 * phi) + 4 * I * (
            36 * omega**2 + 6 * omega + 1
        ) / (A * L * phi)
        D = -L * (360 * omega**2 + 60 * omega + 1) / (30 * phi) - 2 * I * (
            72 * omega**2 + 12 * omega - 1
        ) / (A * L * phi)

    k = (
        np.block(
            [
                [N, N, N1, N, N, N2],
                [N, A, B, N, -A, B],
                [N1, B, C, -N1, -B, D],
                [N, N, -N1, N, N, -N2],
                [N, -A, -B, N, A, -B],
                [N2, B, D, -N2, -B, C],
            ]
        )
        * -P
    )

    return k


@njit(parallel=True, cache=True)
def jit_build(k, ndim, n_elem, node_per_element, ndof_per_node):
    # pre-allocate stiffness matrix
    K = np.zeros((ndim, ndim), dtype=np.float64)

    for i in prange(n_elem):
        # dummy stiffness matrix that updates at each iteration
        K_temp = np.zeros((ndim, ndim), dtype=np.float64)
        # select relevant rows/columns of dummy stiffness matrix
        start = ndof_per_node * i
        end = ndof_per_node * i + ndof_per_node * node_per_element
        K_temp[start:end, start:end] = k[i]

        # update global stiffness matrix
        K += K_temp

    return K


def build_stiffness_matrix(model, u=None, kind=None):
    """Builds the stiffness matrix based on the model(element and node) properties

    Element stiffness matrices are first computed for each element and then loaded in the global stiffness matrix through summation.
    Different element stiffness matrices are computed depending on the specifications found in the Model object.

    Parameters
    ----------
    model : obj from openpile library
        stores all information about nodes elements, and other model-related items, see #TODO:link to model class
    f: np.ndarray, Optional
        Global internal force vector.
    u: np.ndarray, Optional
        Global displacement vector. Must be given if soil_flag is not None
    p_mobilised: np.ndarray, Optional
        Mobilised p-value. Must be given if soil_flag is not None
    kind: str, Optional
        "initial", "secant" or "tangent". Must be given if soil_flag is not None

    Returns
    -------
    K : numpy array of dim (ndof, ndof)
        Global stiffness matrix with all dofs in unit:1/kPa.
    """

    # number of dof per node
    ndof_per_node = 3
    # number of nodes per element
    node_per_element = 2
    # number of elements in mesh
    n_elem = model.element_number
    # global dimension of the global stiffness matrix can be inferred
    ndim_global = ndof_per_node * n_elem + ndof_per_node

    # mechanical stiffness properties
    k = elem_mechanical_stiffness_matrix(model)
    # k += elem_p_delta_stiffness_matrix(model, u)

    # add soil contribution
    if model.soil is not None:
        # gives warning if soil is given without displacements or type of stiffness
        if u is None or kind is None:
            UserWarning("'u' and 'kind' must be stipulated.")
        else:
            if model.distributed_lateral:
                k += elem_py_stiffness_matrix(model, u, kind)

            if model.distributed_axial:
                k += elem_tz_stiffness_matrix(model, u, kind)

            if model.distributed_moment:
                if not model.distributed_lateral:
                    raise UserInputError(
                        "Error: Distributed moment cannot be calculated without distributed lateral springs."
                    )
                k += elem_mt_stiffness_matrix(model, u, kind)

    K = jit_build(k, ndim_global, n_elem, node_per_element, ndof_per_node)

    # add base springs contribution
    if model.soil is not None:
        if model.base_shear:
            K[-2, -2] += calculate_base_spring_stiffness(u[-2], model._Hb_spring, kind)

        if model.base_moment:
            K[-1, -1] += calculate_base_spring_stiffness(u[-1], model._Mb_spring, kind)

        if model.base_axial:
            K[-3, -3] += calculate_base_axial_spring_stiffness(u[-3], model._qz_spring, kind)

    return K


def mesh_to_global_force_dof_vector(df: pd.DataFrame) -> np.ndarray:
    # extract each column (one line per node)
    force_dof_vector = df[["Pz [kN]", "Py [kN]", "Mx [kNm]"]].values.reshape(-1).astype(np.float64)

    return force_dof_vector


def mesh_to_global_disp_dof_vector(df: pd.DataFrame) -> np.ndarray:
    # extract each column (one line per node)
    disp_dof_vector = df[["Tz [m]", "Ty [m]", "Rx [rad]"]].values.reshape(-1).astype(np.float64)

    return disp_dof_vector


def mesh_to_global_restrained_dof_vector(df: pd.DataFrame) -> np.ndarray:
    # extract each column (one line per node)
    restrained_dof_vector = df[["Tz", "Ty", "Rx"]].values.reshape(-1)

    return restrained_dof_vector


def pile_internal_forces(model, u):
    # number of dof per node
    ndof_per_node = 3
    # number of nodes per element
    node_per_element = 2

    # create mech consistent stiffness matrix
    k = elem_mechanical_stiffness_matrix(model)

    # create array u of shape [n_elem x 6 x 1]
    u = global_dof_vector_to_consistent_stacked_array(u, ndof_per_node * node_per_element)
    # compute internal forces and reshape into global dof vector
    F_int = (-1) * np.matmul(k, u).reshape((-1))

    return F_int


@njit(cache=True)
def calculate_base_spring_stiffness(
    u: np.ndarray, spring: np.ndarray, kind: Literal["initial", "secant", "tangent"]
):
    """Calculate springs stiffness for py or t-z springs.

    Parameters
    ----------
    u : float
        base displacement or rotation to calculate stiffness
    spring : np.ndarray
        soil-structure interaction base springs array of shape (1, 1, 1, spring_dim)
    kind : str
        defines whether it is initial, secant of tangent stiffness to define

    Returns
    -------
    k: float
        secant or tangent stiffness for all elements.
    """

    # displacemet with same dimension as spring
    d = abs(u)

    if np.sum(spring[0, 0, 0]) == 0:
        k = 0.0
    else:
        if kind == "initial" or d == 0.0:
            dx = spring[0, 0, 1, 1] - spring[0, 0, 1, 0]
            p0 = spring[0, 0, 0, 0]
            p1 = spring[0, 0, 0, 1]
        elif kind == "secant":
            dx = d
            p0 = spring[0, 0, 0, 0]
            if d > np.max(spring[0, 0, 1]):
                p1 = spring[0, 0, 0, -1]
            else:
                p1 = np.interp(dx, spring[0, 0, 1], spring[0, 0, 0])
        elif kind == "tangent":
            dx = min(0.0005, d)
            if (d - dx) > np.max(spring[0, 0, 1]):
                p0 = spring[0, 0, 0, -1]
            else:
                p0 = np.interp(d - dx, spring[0, 0, 1], spring[0, 0, 0])
            if d > np.max(spring[0, 0, 1]):
                p1 = spring[0, 0, 0, -1]
            else:
                p1 = np.interp(d, spring[0, 0, 1], spring[0, 0, 0])

        k = abs((p1 - p0) / dx)

    return k


@njit(cache=True)
def calculate_base_axial_spring_stiffness(
    u: np.ndarray, spring: np.ndarray, kind: Literal["initial", "secant", "tangent"]
):
    """Calculate springs stiffness for base axial spring.

    Parameters
    ----------
    u : float
        base displacement or rotation to calculate stiffness
    spring : np.ndarray
        soil-structure interaction base springs array of shape (1, 1, 1, spring_dim)
    kind : str
        defines whether it is initial, secant of tangent stiffness to define

    Returns
    -------
    k: float
        secant or tangent stiffness for all elements.
    """

    # displacement with same dimension as spring
    d = u

    # determine place where 0 value of qz-spring is located
    # TODO: make this dynamic
    spring_0_index = round(spring.shape[-1] / 2) - 1

    if np.sum(np.abs(spring[0, 0, 0])) == 0:
        k = 0.0
    else:
        if kind == "initial" or d == 0.0:
            dx = spring[0, 0, 1, spring_0_index] - spring[0, 0, 1, spring_0_index + 1]
            p0 = spring[0, 0, 0, spring_0_index]
            p1 = spring[0, 0, 0, spring_0_index + 1]
        elif kind == "secant":
            dx = d
            p0 = spring[0, 0, 0, spring_0_index]
            p1 = np.interp(dx, spring[0, 0, 1][::-1], spring[0, 0, 0][::-1])
        elif kind == "tangent":
            dx = min(0.0005, abs(d))
            if d > 0:
                p0 = np.interp(d - dx, spring[0, 0, 1][::-1], spring[0, 0, 0][::-1])
            else:
                p0 = np.interp(d + dx, spring[0, 0, 1][::-1], spring[0, 0, 0][::-1])

            p1 = np.interp(d, spring[0, 0, 1][::-1], spring[0, 0, 0][::-1])

        k = abs((p1 - p0) / dx)

    return k


@njit(cache=True)
def calculate_py_springs_stiffness(
    u: np.ndarray, springs: np.ndarray, kind: Literal["initial", "secant", "tangent"]
):
    """Calculate springs stiffness for py springs.

    Parameters
    ----------
    u : np.ndarray
        displacements to calculate stiffness.
        For dofs related to t-z curves, u = U[::3] where U is the global displacement vector.
        For dofs related to p-y curves, u = U[1::3] where U is the global displacement vector.
    springs : np.ndarray
        soil-structure interaction py springs array of shape (n_elem, 2, 2, spring_dim)
    kind : str
        defines whether it is initial, secant of tangent stiffness to define

    Returns
    -------
    k: np.ndarray
        secant or tangent stiffness for all elements. Array of shape(n_elem,2,1,1)
    """

    # double inner values for u
    d = double_inner_njit(u)

    # displacemet with same dimension as spring
    d = np.abs(d).reshape((-1, 2, 1, 1))

    k = np.zeros(d.shape, dtype=np.float64)

    for i in range(k.shape[0]):
        for j in range(k.shape[1]):
            if np.sum(springs[i, j, 1]) == 0:
                pass
            else:
                if kind == "initial" or d[i, j, 0, 0] == 0.0:
                    dx = springs[i, j, 1, 1] - springs[i, j, 1, 0]
                    p0 = springs[i, j, 0, 0]
                    p1 = springs[i, j, 0, 1]
                elif kind == "secant":
                    dx = d[i, j, 0, 0]
                    p0 = springs[i, j, 0, 0]
                    if d[i, j, 0, 0] > np.max(springs[i, j, 1]):
                        p1 = springs[i, j, 0, -1]
                    else:
                        p1 = np.interp(dx, springs[i, j, 1], springs[i, j, 0])
                elif kind == "tangent":
                    dx = min(0.0005, d[i, j, 0, 0])
                    if (d[i, j, 0, 0] - dx) > np.max(springs[i, j, 1]):
                        p0 = springs[i, j, 0, -1]
                    else:
                        p0 = np.interp(d[i, j, 0, 0] - dx, springs[i, j, 1], springs[i, j, 0])
                    if d[i, j, 0, 0] > np.max(springs[i, j, 1]):
                        p1 = springs[i, j, 0, -1]
                    else:
                        p1 = np.interp(d[i, j, 0, 0], springs[i, j, 1], springs[i, j, 0])

                k[i, j, 0, 0] = abs((p1 - p0) / dx)

    return k


# @njit(cache=True)
def calculate_tz_springs_stiffness(
    u: np.ndarray, springs: np.ndarray, kind: Literal["initial", "secant", "tangent"]
):
    """Calculate springs stiffness for t-z springs.

    Parameters
    ----------
    u : np.ndarray
        displacements to calculate stiffness.
        For dofs related to t-z curves, u = U[::3] where U is the global displacement vector.
        For dofs related to p-y curves, u = U[1::3] where U is the global displacement vector.
    springs : np.ndarray
        soil-structure interaction py springs array of shape (n_elem, 2, 2, spring_dim)
    kind : str
        defines whether it is initial, secant of tangent stiffness to define

    Returns
    -------
    k: np.ndarray
        secant or tangent stiffness for all elements. Array of shape(n_elem,2,1,1)
    """

    # double inner values for u
    d = double_inner_njit(u)

    # displacement with same dimension as spring
    d = np.array(d).reshape((-1, 2, 1, 1))

    k = np.zeros(d.shape, dtype=np.float64)

    # determine place where 0 value of tz-spring is located
    spring_0_index = round(springs.shape[-1] / 2) - 1

    for i in range(k.shape[0]):
        for j in range(k.shape[1]):
            if np.sum(np.abs(springs[i, j, 1])) == 0:
                pass
            else:
                if kind == "initial" or d[i, j, 0, 0] == 0.0:
                    dx = springs[i, j, 1, spring_0_index] - springs[i, j, 1, spring_0_index + 1]
                    p0 = springs[i, j, 0, spring_0_index]
                    p1 = springs[i, j, 0, spring_0_index + 1]
                elif kind == "secant":
                    dx = d[i, j, 0, 0]
                    p0 = springs[i, j, 0, spring_0_index]
                    p1 = np.interp(dx, springs[i, j, 1][::-1], springs[i, j, 0][::-1])
                elif kind == "tangent":
                    dx = min(0.0005, abs(d[i, j, 0, 0]))
                    if d[i, j, 0, 0] > 0:
                        p0 = np.interp(
                            d[i, j, 0, 0] - dx, springs[i, j, 1][::-1], springs[i, j, 0][::-1]
                        )
                    else:
                        p0 = np.interp(
                            d[i, j, 0, 0] + dx, springs[i, j, 1][::-1], springs[i, j, 0][::-1]
                        )

                    p1 = np.interp(d[i, j, 0, 0], springs[i, j, 1][::-1], springs[i, j, 0][::-1])

                k[i, j, 0, 0] = abs((p1 - p0) / dx)

    return k


@njit(cache=True)
def calculate_mt_springs_stiffness(
    u: np.ndarray,
    mt_springs: np.ndarray,
    py_springs: np.ndarray,
    p_mobilised: np.ndarray,
    kind: Literal["initial", "secant", "tangent"],
):
    """Calculate springs stiffness for rotational springs

    .. note::
        The difference with the py function is that rotational springs can depend on lateral springs

    Parameters
    ----------
    u : np.ndarray
        displacements to calculate stiffness.
        For dofs related to m-t curves, u = U[2::3] where U is the global displacement vector.
    mt_springs : np.ndarray
        soil-structure interaction m-t springs array of shape (n_elem, 2, 2, py_spring_dim, mt_spring_dim)
    py_springs : np.ndarray
        soil-structure interaction p-y springs array of shape (n_elem, 2, 2, py_spring_dim)
    p_mobilised : np.ndarray
        current p value of p-y springs of shape (nelem, 2, 1, 1)
    kind : str
        defines whether it is initial, secant of tangent stiffness to define

    Returns
    -------
    k: np.ndarray
        secant or tangent stiffness for all elements. Array of shape(n_elem,2,1,1)
    """

    # double inner values for u
    d = double_inner_njit(u)

    # displacemet and p_mobilised with same dimension as spring
    d = np.abs(d).reshape((-1, 2, 1, 1))
    p = p_mobilised.reshape((-1, 2, 1, 1))

    k = np.zeros(d.shape, dtype=np.float64)

    # get the proper m-t spring
    m = np.zeros(mt_springs.shape[4])
    t = np.zeros(mt_springs.shape[4])

    # i for each element
    for i in range(k.shape[0]):
        # j for top and bottom values of spring
        for j in range(k.shape[1]):
            if np.sum(mt_springs[i, j, 1, 0]) == 0:
                # check if first m-vector is not a defined spring
                # if that is the case, we do not calculate any stiffness
                pass
            else:

                for ii in range((mt_springs.shape[4])):
                    if ii == 0:
                        pass
                    else:
                        m[ii] = np.interp(
                            p[i, j, 0, 0], py_springs[i, j, 0, :], mt_springs[i, j, 0, :, ii]
                        )
                        t[ii] = np.interp(
                            p[i, j, 0, 0], py_springs[i, j, 0, :], mt_springs[i, j, 1, :, ii]
                        )

                if kind == "initial" or d[i, j, 0, 0] == 0.0:
                    dt = t[1] - t[0]
                    m0 = m[0]
                    m1 = m[1]
                elif kind == "secant":
                    dt = d[i, j, 0, 0]
                    m0 = m[0]
                    if d[i, j, 0, 0] > t[-1]:
                        m1 = m[-1]
                    else:
                        m1 = np.interp(dt, t, m)
                elif kind == "tangent":
                    dt = min(0.01 * t[1], d[i, j, 0, 0])
                    if (d[i, j, 0, 0] - dt) > t[-1]:
                        m0 = m[-1]
                    else:
                        m0 = np.interp(d[i, j, 0, 0] - dt, t, m)
                    if (d[i, j, 0, 0]) > t[-1]:
                        m1 = m[-1]
                    else:
                        m1 = np.interp(d[i, j, 0, 0], t, m)

                k[i, j, 0, 0] = abs(m1 - m0) / (dt)

    return k


def computer():
    """This function is the solver of openpile.

    The solver reads the boundary conditions, the global Force and Displacement vector and iterate to find convergence.
    """



# Import libraries
import math as m
import numpy as np
from numba import njit

# maximum resistance values
@njit(cache=True)
def _Qmax_api_clay(
    Su: float,
) -> float:
    # Unit end-bearing [kPa]
    return 9 * Su


@njit(cache=True)
def _Qmax_api_sand(
    sig: float,
    delta: float,
) -> float:
    # important variables
    delta_table = np.array([0, 15, 20, 25, 30, 35, 100], dtype=np.float32)
    Nq_table = np.array([8, 8, 12, 20, 40, 50, 50], dtype=np.float32)
    Qmax_table = np.array([1900, 1900, 2900, 4800, 9600, 12000, 12000], dtype=np.float32)

    Nq = np.interp(delta, delta_table, Nq_table)
    Qmax = np.interp(delta, delta_table, Qmax_table)

    # Unit end-bearing [kPa]
    return min(Qmax, sig * Nq)


@njit(cache=True)
def _fmax_api_clay(
    sig: float,
    Su: float,
    alpha_limit: float,
) -> float:
    """Creates the maximum skin friction.

    The methdology follows the API clay method of axial capacity found in .

    Parameters
    ----------
    sig : float
        vertical effcitve stress in kPa.
    Su : float
        undrained shear strength in kPa.
    alpha_limit : float
        limit value for the skin friction normalized to undrained shear strength.


    Returns
    -------
    float
        unit skin friction in kPa.
    """

    # important variables
    if sig == 0.0:
        psi = Su / 0.001
    else:
        psi = Su / sig

    if psi > 1.0:
        alpha = min(0.5 * psi ** (-0.25), alpha_limit)
    else:
        alpha = min(0.5 * psi ** (-0.5), alpha_limit)

    # Unit skin friction [kPa]
    return alpha * Su


# SPRING FUNCTIONS --------------------------------------------

# API sand function
@njit(cache=True)
def _fmax_api_sand(
    sig: float,
    delta: float,
    K: float = 0.8,
) -> float:
    """Creates the maximum skin friction.

    The methdology follows the API sand method of axial capacity found in .

    Parameters
    ----------
    sig : float
        vertical effcitve stress in kPa.
    delta: float
        interface friction angle in degrees
    K: float
        coefficient of lateral pressure.
        (0.8 for open-ended piles and 1.0 for cloased-ended)

    Returns
    -------
    float
        unit skin friction in kPa.
    """

    # important variables
    delta_table = np.array([0, 15, 20, 25, 30, 35, 100], dtype=np.float32)
    fs_max_table = np.array([47.8, 47.8, 67, 81.3, 95.7, 114.8, 114.8], dtype=np.float32)

    # limit unit skin friction according to API ref page 59
    fs_max = np.interp(delta, delta_table, fs_max_table)

    # Unit skin friction [kPa]
    return min(fs_max, K * sig * m.tan(delta * m.pi / 180.0))
